"""Algorithms to work with 1D signals."""

import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks
from scipy.special import erfc
from scipy.stats import median_abs_deviation as mad

from ..core.utils.numpy import FloatArray1D, IntArray1D


def detect_peaks(
    x: np.ndarray, noise: np.ndarray, baseline: np.ndarray, **kwargs
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""Find peaks in a 1D signal.

    :param x: 1D array.
    :param noise: the noise level at each point in `x`. **MUST** have the same size as `x`.
    :param baseline: the baseline level at each point in `x`. **MUST** have the same size as `x`.
    :param kwargs: extra parameters to pass to :py:func:`scipy.signal.find_peaks`. If the `prominence`
        parameter is passed, it will be ignored and set to three times the noise level.
    :return: a tuple consisting of: an int array with peaks start location, an int array
        with the apexes location and an int array with the peaks end location.

    Algorithm
    ---------

    1.  Peak apexes are detected using :py:func:`scipy.signal.find_peaks` with a minimum
        distance of three points and a minimum prominence equal to three times the noise level.
    2.  points in :math:`x` are classified as either signal or baseline. The k-th point in :math:`x` is
        classified as baseline if the following condition is met:

        .. math::
            |x[k] - b[k]| < e[k]

        where :math:`b` is the baseline and :math:`e` is the noise.
    3.  Peaks are removed if they fall in a in a region classified as baseline.
    4.  Peak extensions, i.e., beginning and end, are defined as the closest baseline point to
        the left and right of each apex.
    5.  Overlapping peak extensions are fixed by setting the boundary between the peaks to the minimum
        value between the two apexes.

    .. seealso::
        - :py:func:`estimate_noise`: Estimate the noise level in a 1D signal.
        - :py:func:`estimate_baseline`: Estimate the baseline in a 1D signal.

    """
    baseline_index = np.where((x - baseline) < noise)[0]
    if baseline_index.size == 0:
        baseline_index = np.array([0, x.size - 1], dtype=int)

    prominence = 3 * noise

    if kwargs:
        kwargs["prominence"] = prominence
    else:
        kwargs.update({"prominence": prominence, "distance": 3})

    peaks = find_peaks(x, **kwargs)[0]

    peaks = np.setdiff1d(peaks, baseline_index, assume_unique=True)

    start, end = _find_peak_extension(peaks, baseline_index)
    start, end = _fix_peak_extension_overlap(x, start, peaks, end)
    start, peaks, end = _normalize_peaks(x, start, peaks, end)
    return start, peaks, end


def estimate_noise(x: FloatArray1D, n_chunks: int = 5, robust: bool = True, min_chunk_size: int = 200) -> np.ndarray:
    """Estimate the noise level in a 1D signal.

    `x` is split into equally sized chunks and a noise estimation is done assuming a gaussian
    iid in each chunk. See [ADD LINK] for a detailed description of how the method
    works.

    :param x: a 1D array
    :param n_chunks: number of chunks to create. The size of each slice must be greater than
        `min_chunk_size`.
    :param robust: if set to ``True``, estimates the noise using the median absolute deviation. Otherwise,
        noise estimation uses the standard deviation.
    :param min_slice_size: minimum size of a slice. If the size of x is smaller than this value,
        the noise is estimated using the whole array.
    :return: an array that contains the noise level at each point in `x`.

    """
    noise = np.zeros_like(x)
    slice_size = x.size // n_chunks
    if slice_size < min_chunk_size:
        slice_size = min_chunk_size
    start = 0
    while start < x.size:
        end = min(start + slice_size, x.size)

        # prevent short slices at the end of x
        if (x.size - end) < min_chunk_size:
            end = x.size

        slice_noise = _estimate_local_noise(x[start:end], robust=robust)
        noise[start:end] = slice_noise
        start = end
    return noise


def estimate_baseline(x: np.ndarray, noise: np.ndarray, min_proba: float = 0.05) -> np.ndarray:
    """Estimate the baseline level of a 1D signal.

    :param x: non-empty 1D array
    :param noise: the noise level at each point in `x`. **MUST** have the same size as `x`.
    :param min_proba: number between 0 and 1, default=0.05
    :return: an array that contains the baseline level at each point in `x`.

    Algorithm
    ---------

    The baseline is estimated by classifying each point in the signal as either
    signal or baseline. The baseline is obtained by interpolation of baseline
    points. See [ADD LINK] for a detailed explanation of how the method works.

    .. seealso::
        - :py:func:`estimate_noise`: Estimate the noise level in a 1D signal.

    """
    # find points that only have contribution from the baseline
    baseline_index = _find_baseline_points(x, noise, min_proba)

    # interpolate baseline points to match x size
    baseline = x[baseline_index]
    interpolator = interp1d(baseline_index, baseline)
    baseline = interpolator(np.arange(x.size))

    # prevents that interpolated points have higher values than x.
    baseline = np.minimum(baseline, x)
    return baseline


def find_centroids(
    mz: np.ndarray, spint: np.ndarray, min_snr: float, min_distance: float
) -> tuple[FloatArray1D, FloatArray1D]:
    """Find the centroid of a mass spectrum in profile mode.

    :param mz: array of m/z in profile mode.
    :param spint: array of spectral intensity in profile mode.
    :param min_snr: minimum peak signal-to-noise ratio
    :param min_distance: minimum m/z distance between consecutive centroids

    Returns
    -------
    centroid_mz : array
        centroid m/z of peaks
    centroid_int : array
        area of peaks

    """
    noise = estimate_noise(spint)
    baseline = estimate_baseline(spint, noise)
    baseline_index = np.where((spint - baseline) < noise)[0]
    prominence = 3 * noise
    find_peaks_params = {"prominence": prominence, "distance": 3}

    peaks = find_peaks(spint, **find_peaks_params)[0]
    # remove peaks close to baseline level
    peaks = np.setdiff1d(peaks, baseline_index, assume_unique=True)
    peaks = peaks[((spint[peaks] - baseline[peaks]) / noise[peaks]) > min_snr]

    start, end = _find_peak_extension(peaks, baseline_index)
    start, end = _fix_peak_extension_overlap(spint, start, peaks, end)
    start, peaks, end = _normalize_peaks(spint, start, peaks, end)

    # peak centroid and total intensity computation
    # if m[0], ...,  m[n] is the m/z array and i[0], ..., i[n] is the
    # intensity array, for a peak with start and end indices k and l respectively,
    # the total intensity A is A = \sum_{j=k}^{l} i[j] and the centroid C,
    # computed as the weighted mean of the m/z is
    # C = \sum_{j=k}^{l} m[j] * i[j] / A
    # If we define the cumulative intensity I_{k} = \sum_{j=0}^{k} i[j]
    # It is easy to see that A = I[l - 1] - I[k - 1]. The same can be done
    # for the centroids defining the weights W[k] = \sum_{j=0}^{k} m[j] * i[j]
    # C = (W[l - 1] - W[k - 1]) / A
    if start.size:
        cumulative_spint = np.cumsum(spint)
        weights = np.cumsum(mz * spint)
        start_cumulative_spint = cumulative_spint[start - 1]
        if start[0] == 0:
            # prevents using the last value from cumulative_spint
            start_cumulative_spint[0] = 0
        total_spint = cumulative_spint[end - 1] - start_cumulative_spint

        start_weight = weights[start - 1]
        if start[0] == 0:
            start_weight[0] = 0
        centroid = (weights[end - 1] - start_weight) / total_spint
    else:
        centroid = np.array([])
        total_spint = np.array([])

    if centroid.size:
        _merge_close_peaks(centroid, total_spint, min_distance)
    return centroid, total_spint


def smooth(x: FloatArray1D, smoothing_strength: float):
    """Smooth a signal using using a gaussian kernel.

    :param smoothing_strength : standard deviation of the gaussian kernel.
    """
    return gaussian_filter1d(x, smoothing_strength)


def _find_peak_extension(peaks: IntArray1D, baseline_index: IntArray1D) -> tuple[IntArray1D, IntArray1D]:
    """Find the closest baseline points to the left and right of each peak.

    aux function of detect_peaks
    peaks : array of apex indices
    baseline_index : array of indices
    :return: a tuple with arrays of peak beginning and end

    """
    ext_index = np.searchsorted(baseline_index, peaks)
    ext_index[ext_index >= baseline_index.size] = baseline_index.size - 1
    start = baseline_index[ext_index - 1]
    end = baseline_index[ext_index] + 1
    return start, end


def _fix_peak_extension_overlap(
    y: np.ndarray, start: np.ndarray, peaks: np.ndarray, end: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    local_min = find_peaks(-y)[0]
    # find overlapping peaks indices
    overlap_mask = end > np.roll(start, -1)
    if overlap_mask.size:
        overlap_mask[-1] = False
    overlap_index = np.where(overlap_mask)[0]
    for k in overlap_index:
        # search local min in the region defined by the overlapping peaks
        ks, ke = np.searchsorted(local_min, [peaks[k], peaks[k + 1]])
        k_min = np.argmin(y[local_min[ks:ke]])
        boundary = local_min[ks + k_min]
        end[k] = boundary + 1
        start[k + 1] = boundary
    return start, end


def _normalize_peaks(
    y: np.ndarray, start: np.ndarray, peaks: np.ndarray, end: np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Perform sanity check on peaks.

    Finds the local maximum in the region defined between start and end for each
    peak. If no maximum is found, the peak is removed. Also, corrects the peak
    index using this value that can be slightly shifted if smoothing was
    applied to the signal.

    aux function of detect_peaks.

    """
    local_max = find_peaks(y)[0]
    start_index = np.searchsorted(local_max, start)
    end_index = np.searchsorted(local_max, end, side="right")
    fixed_peaks = np.zeros_like(peaks)
    valid_peaks = np.zeros(peaks.size, dtype=bool)
    for k in range(peaks.size):
        local_max_slice = y[local_max[start_index[k] : end_index[k]]]
        if local_max_slice.size > 0:
            k_max = np.argmax(local_max_slice)
            new_peak = local_max[start_index[k] + k_max]
            valid_peaks[k] = (start[k] < new_peak) and (new_peak < end[k])
            fixed_peaks[k] = new_peak
    start = start[valid_peaks]
    end = end[valid_peaks]
    fixed_peaks = fixed_peaks[valid_peaks]
    return start, fixed_peaks, end


def _estimate_local_noise(x: FloatArray1D, robust: bool = True) -> float:
    r"""Estimate noise in a 1D signal assuming that the noise is gaussian iid.

    :param x: 1D array with a length greater or equal than four.If the size is smaller, the
        function will return 0.
    :param robust: if set to ``True``, estimates the noise using the median absolute deviation.
        Otherwise, uses the standard deviation.

    """
    # if d2x follows a normal distribution ~ N(0, 2*sigma), its sample mean
    # has a normal distribution ~ N(0,  2 * sigma / sqrt(n - 2)) where n is the
    # size of d2x.
    # d2x with high absolute values are removed until this the mean of d2x is
    # lower than its standard deviation.
    # start at 90th percentile and decrease the percentile by 10 in each iteration.
    # The loop stops at the 20th percentile even if this condition is not meet

    d2x = np.diff(x, n=2)
    d2x = d2x[np.argsort(np.abs(d2x))]

    # sizes at 20, 30, ..., 90 percentiles
    # sizes must be > 2 to compute the standard deviation
    chunk_sizes = [d2x.size * percentile // 100 for percentile in reversed(range(20, 100, 10))]

    for size in chunk_sizes:
        if size <= 2:
            break

        noise_std: float = mad(d2x[:size], scale="normal") if robust else np.std(d2x[:size]).item()  # type: ignore

        # if all the values in d2x are equal, the noise level is zero
        if noise_std == 0.0:
            return 0.0

        noise_mean = np.median(d2x[:size]).item() if robust else np.mean(d2x[:size]).item()

        if abs(noise_mean / noise_std) <= 1.0:
            return noise_std / 2

    # if there are no chunk with sizes greater than 2, fall back to zero
    return 0.0


def _find_baseline_points(x: np.ndarray, noise: np.ndarray, min_proba: float) -> np.ndarray:
    """Find points flagged as baseline.

    Aux function of estimate_baseline.

    """
    extrema = _find_local_extrema(x)
    # check how likely is that the difference observed in each min-max slice
    # can be attributed to noise.
    noise_proba = _estimate_noise_probability(noise, x, extrema)
    # creates a vector with indices where baseline was found
    baseline_index = _build_baseline_index(x, noise_proba, min_proba, extrema)
    return baseline_index


def _find_local_extrema(x: np.ndarray) -> np.ndarray:
    """Find all local minima and maxima in an 1D array.

    aux function of _find_baseline_points.

    """
    local_max = find_peaks(x)[0]
    local_min = find_peaks(-x)[0]
    if local_max.size:
        # include first and last indices
        extrema = np.hstack([0, local_min, local_max, x.size - 1])
    else:
        extrema = np.array([], dtype=int)
    return np.unique(extrema)


def _estimate_noise_probability(noise: np.ndarray, x: np.ndarray, extrema: np.ndarray) -> np.ndarray:
    """Compute the probability that the variation observed in each slice is due to noise only.

    Aux function of _find_baseline_points.

    """
    if extrema.size:
        noise_slice = _get_noise_slice_sum_std(noise, extrema[:-1])
        # The difference between maximum and minimum in each slice
        # delta = np.abs(x[np.roll(extrema, -1)] - x[extrema])[:-1]
        x_sum = _get_signal_sum(x, extrema)
        # here we are computing p(abs(sum noise) > delta) assuming a normal
        # distribution
        noise_probability = erfc(x_sum / (noise_slice * np.sqrt(2)))
    else:
        # prevents raising an exception when no extrema had been found
        noise_probability = np.array([])
    return noise_probability


def _get_noise_slice_sum_std(noise: np.ndarray, extrema: np.ndarray) -> np.ndarray:
    """Compute the standard deviation of the sum of slices between local maxima.

    aux function of _estimate_noise_probability.

    """
    # the values in noise are an estimation of the standard deviation of the
    # noise. If the noise is iid, the std of the sum is the sum of variances.

    # reshape the extrema indices to compute the sum of elements between
    # consecutive slices, i.e: the sum between of y between extrema[0] and
    # extrema[1], extrema[1] and extrema[2]...
    # The last element is not used
    reduce_ind = np.vstack([extrema, np.roll(extrema + 1, -1)]).T.reshape(extrema.size * 2)[:-1]
    return np.sqrt(np.add.reduceat(noise**2, reduce_ind)[::2])


def _get_signal_sum(x, extrema):
    # we want to compute the sum of x between a, b, where a and b two
    # consecutive local maxima indices. if X is the cumulative sum of x, then
    # the sum between a and b is X[b] - X[a - 1]. The sum must also be relative
    # to the minimum between a and b. Because we are comparing intervals
    # between local extrema, the intervals are monotonic, and we just need to
    # subtract the min(x[a], x[b]) multiplied by the length of the interval.
    # this code achieves this in a vectorized way
    cum_x = np.cumsum(x)
    ext_shift = np.roll(extrema, -1)
    n_times = ext_shift - extrema + 1
    start_cum_int = cum_x[extrema - 1]
    # first value set to 0 to avoid errors associated with roll
    start_cum_int[0] = 0
    x_sum = cum_x[ext_shift] - start_cum_int
    x_min = np.minimum(x[ext_shift], x[extrema])
    res = x_sum - n_times * x_min
    res = res[:-1]
    return res


def _build_baseline_index(x: np.ndarray, noise_probability: np.ndarray, min_p: float, extrema: np.ndarray):
    """Build an array with indices of points flagged as baseline.

    aux function of _find_baseline_points

    """
    # define regions of signal based on noise probability
    is_signal = noise_probability < min_p
    # extend regions of signals to right and left
    is_signal = is_signal | np.roll(is_signal, 1) | np.roll(is_signal, -1)
    baseline_index = list()

    for k in range(extrema.size - 1):
        if not is_signal[k]:
            slice_indices = np.arange(extrema[k], extrema[k + 1] + 1)
            baseline_index.append(slice_indices)

    baseline_index = _include_first_and_last_index(x, baseline_index)
    return baseline_index


def _include_first_and_last_index(x: FloatArray1D, fixed_baseline_index: list[IntArray1D]) -> IntArray1D:
    """Add first and last indices of x to the baseline indices.

    aux function of _build_baseline_index
    """
    if len(fixed_baseline_index):
        stack = list()
        # include first and last indices
        if fixed_baseline_index[0][0] != 0:
            stack.append([0])

        stack.extend(fixed_baseline_index)

        if fixed_baseline_index[-1][-1] != x.size - 1:
            stack.append([x.size - 1])

        fixed_baseline_index = np.hstack(stack)  # type: ignore
    else:
        fixed_baseline_index = np.array([0, x.size - 1], dtype=int)  # type: ignore
    return fixed_baseline_index  # type: ignore


def _merge_close_peaks(mz: np.ndarray, spint: np.ndarray, min_distance: float) -> tuple[np.ndarray, np.ndarray]:
    if mz.shape[0] < 2:
        return mz, spint
    dmz = np.diff(mz)
    is_close_mask = (dmz < min_distance) & (np.roll(dmz, -1) > min_distance)
    is_close_mask[-1] = (mz[-1] - mz[-2]) < min_distance  # boundary case
    close_index = np.where(is_close_mask)[0]
    while close_index.size > 0:
        # merge close centroids
        new_spint = spint[close_index] + spint[close_index + 1]
        new_mz = mz[close_index] * spint[close_index] + mz[close_index + 1] * spint[close_index + 1]
        new_mz /= new_spint
        spint[close_index] = new_spint
        mz[close_index] = new_mz
        # remove merged centroids
        mz = np.delete(mz, close_index + 1)
        spint = np.delete(spint, close_index + 1)

        # repeat until no close peaks are detected
        dmz = np.diff(mz)
        is_close_mask = (dmz < min_distance) & (np.roll(dmz, -1) > min_distance)
        is_close_mask[-1] = (mz[-1] - mz[-2]) < min_distance
        close_index = np.where(is_close_mask)[0]
    return mz, spint



"""Data storage classes."""

from .memory import OnMemoryAssayStorage, OnMemorySampleStorage
from .sqlite import SQLiteAssayStorage

__all__ = ["OnMemoryAssayStorage", "OnMemorySampleStorage", "SQLiteAssayStorage"]



import numpy as np
import pytest

from tidyms2.core.models import MSSpectrum
from tidyms2.io import MSData
from tidyms2.simulation import lcms
from tidyms2.simulation.base import DataAcquisitionSpec, MZGridSpec

from ..helpers import create_sample


class TestSimulatedLCMSSample:
    def test_make_grid_no_spec_ok(self, lcms_sample_factory: lcms.SimulatedLCMSSampleFactory):
        features = lcms_sample_factory.create_features()
        simulated_sample_spec = lcms.SimulatedLCMSSample(
            config=lcms_sample_factory.data_acquisition, features=features
        )
        grid = simulated_sample_spec.make_grid()
        assert grid.size == len(simulated_sample_spec.features)
        assert np.all(np.diff(grid) > 0.0)

    def test_make_grid_with_spec(self, lcms_sample_factory: lcms.SimulatedLCMSSampleFactory):
        lcms_sample_factory.data_acquisition.grid = MZGridSpec()
        features = lcms_sample_factory.create_features()
        simulated_sample_spec = lcms.SimulatedLCMSSample(
            config=lcms_sample_factory.data_acquisition, features=features
        )
        grid = simulated_sample_spec.make_grid()
        assert simulated_sample_spec.config.grid is not None
        assert grid.size == simulated_sample_spec.config.grid.size
        assert np.all(np.diff(grid) > 0.0)

    def test_make_grid_no_features_return_empty_array(self):
        config = DataAcquisitionSpec()
        features = list()
        simulated_sample_spec = lcms.SimulatedLCMSSample(config=config, features=features)
        grid = simulated_sample_spec.make_grid()
        assert grid.size == 0


class TestSimulatedMSData:
    @pytest.fixture(scope="class")
    def data(self, lcms_sample_factory: lcms.SimulatedLCMSSampleFactory):
        sample = lcms_sample_factory("sample")
        return MSData(sample)

    def test_get_spectrum(self, data: MSData):
        index = 0
        sp = data.get_spectrum(index)
        assert isinstance(sp, MSSpectrum)
        assert sp.index == index

    def test_get_spectrum_negative_index_raises_value_error(self, data: MSData):
        with pytest.raises(ValueError):
            data.get_spectrum(-1)

    def test_get_spectrum_index_greater_than_size_raises_error(self, data: MSData):
        with pytest.raises(ValueError):
            index = 100000000
            data.get_spectrum(index)

    def test_multiple_calls_to_get_spectrum_retrieves_the_same_instance(self, data: MSData):
        index = 10
        sp1 = data.get_spectrum(index)
        sp2 = data.get_spectrum(index)
        assert sp1 is sp2


class TestSimulateDataMatrix:
    def test_no_samples_raise_error(self, lcms_adducts):
        samples = list()
        with pytest.raises(ValueError):
            lcms.simulate_data_matrix(lcms_adducts, samples)

    def test_no_features_raise_error(self):
        samples = [create_sample(create_file=False)]
        lcms_adducts = list()
        with pytest.raises(ValueError):
            lcms.simulate_data_matrix(lcms_adducts, samples)



"""Tools to process LC-MS-based datasets."""

from __future__ import annotations

from typing import Self, assert_never

import pydantic
from numpy import minimum, zeros_like

from ...algorithms.raw import MakeRoiParameters, make_roi
from ...algorithms.signal import detect_peaks, estimate_baseline, estimate_noise, smooth
from ...core.enums import MSInstrument, Polarity, SeparationMode
from ...core.exceptions import InvalidSeparationMode
from ...core.models import MZTrace, Sample
from ...core.operators.sample import FeatureExtractor, RoiExtractor, RoiTransformer
from ...io import MSData
from ..models import Peak


class LCTraceExtractor(RoiExtractor[MZTrace, Peak], MakeRoiParameters):
    """Extracts regions-of-interest (ROI) from raw data represented as m/z traces.

    Traces are created by connecting values across consecutive scans based on the
    closeness in m/z.

    Refer to the :ref:`processing-lcms-datasets` guide for examples on how to use this operator.
    See the :ref:`algorithms-roi-extraction` for a description of the algorithm used.

    .. seealso:: lcms.MZTrace : Representation of a ROI using m/z traces.
    .. seealso:: lcms.MakeRoiParameters : Parameters used by the ROI extraction algorithm

    """

    def extract_rois(self, sample: Sample) -> list[MZTrace]:
        """Apply ROI extraction to a sample with LC data."""
        ms_data = MSData(sample)
        params = MakeRoiParameters.model_validate(self)
        params.ms_level = sample.ms_level
        params.start_time = sample.start_time
        params.end_time = sample.end_time
        return make_roi(ms_data, params)

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        """Set the processor default parameters.

        :param instrument : the instrument type used in the experimental setup
        :param separation : the LC platform used in the experimental setup
        :param polarity : the MS polarity used in the experiment

        """
        match instrument:
            case MSInstrument.QTOF:
                tolerance = 0.01
            case MSInstrument.ORBITRAP:
                tolerance = 0.005
            case _ as never:
                assert_never(never)

        match separation:
            case SeparationMode.UPLC:
                min_length = 10
            case SeparationMode.HPLC:
                min_length = 20
            case SeparationMode.DART:
                raise InvalidSeparationMode(f"{separation} is not a valid LC mode.")
            case _ as never:
                assert_never(never)
        return cls(max_missing=1, pad=2, min_intensity=250.0, tolerance=tolerance, min_length=min_length)


class LCTraceBaselineEstimator(RoiTransformer[MZTrace, Peak]):
    """Estimate the noise level and baseline in an m/z trace.

    The default values for this filter usually produce good results in most LC
    traces. Do not modify these values unless you know what you are doing.
    See :ref:`here <algorithms-peak-extraction>` for a description of the noise estimation
    and baseline estimation algorithms.

    """

    robust: bool = True
    """If ``True``, use the median absolute deviation as an estimator of the
    noise standard deviation. If ``False``. use the standard deviation."""

    min_slice_size: pydantic.PositiveInt = 200
    """The minimum size of a signal slice for local noise estimation. If the signal
    size is smaller than this value, the noise is estimated using the whole array.
    """

    n_slices: pydantic.PositiveInt = 5
    """Number of slices to create. The size of each slice must be greater than
    `min_slice_size`.
    """

    smoothing_strength: pydantic.PositiveFloat | None = 1.0
    """If specified, apply a temporary gaussian smoothing to the trace intensity.
    This step usually improves baseline estimation.
    """

    min_proba: float = pydantic.Field(default=0.05, le=0.5, gt=0.0)
    """The minimum probability of a signal chunk to be considered as baseline."""

    def transform_roi(self, roi: MZTrace):
        """Add noise and baseline to an LC trace."""
        roi.noise = estimate_noise(
            roi.spint, min_chunk_size=self.min_slice_size, n_chunks=self.n_slices, robust=self.robust
        )
        if self.smoothing_strength is not None:
            spint = smooth(roi.spint, self.smoothing_strength)
        else:
            spint = roi.spint

        roi.baseline = minimum(estimate_baseline(spint, roi.noise, min_proba=self.min_proba), roi.spint)
        return roi

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        """Set the processor default parameters.

        :param instrument : the instrument type used in the experimental setup
        :param separation : the LC platform used in the experimental setup
        :param polarity : the MS polarity used in the experiment

        """
        return cls()


class LCTraceSmoother(RoiTransformer[MZTrace, Peak]):
    """Smooth LC traces intensity using a gaussian kernel."""

    strength: pydantic.PositiveFloat = 1.0
    """The smoothing strength, defined as the standard deviation of the gaussian kernel"""

    def transform_roi(self, roi: MZTrace):
        """Add noise and baseline to an LC trace."""
        roi.spint = smooth(roi.spint, self.strength)
        return roi

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        """Set the processor default parameters.

        :param instrument : the instrument type used in the experimental setup
        :param separation : the LC platform used in the experimental setup
        :param polarity : the MS polarity used in the experiment

        """
        return cls()


class LCPeakExtractor(FeatureExtractor[MZTrace, Peak]):
    """Extract peaks from LC m/z traces.

    A complete description of the algorithm used for peak extraction can be found
    :ref:`here <algorithms-peak-extraction>`.

    """

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity):
        """Set the processor default parameters.

        :param instrument : the instrument type used in the experimental setup
        :param separation : the LC platform used in the experimental setup
        :param polarity : the MS polarity used in the experiment

        """
        match separation:
            case SeparationMode.HPLC:
                bounds = {"width": (10, 90), "snr": (5, None)}
            case SeparationMode.UPLC:
                bounds = {"width": (4, 60), "snr": (5, None)}
            case SeparationMode.DART:
                raise NotImplementedError
            case _ as never:
                assert_never(never)
        return cls(bounds=bounds)

    def extract_features(self, roi: MZTrace):
        """Detect peaks in an LC trace."""
        if roi.noise is None:
            roi.noise = estimate_noise(roi.spint)

        if roi.baseline is None:
            roi.baseline = zeros_like(roi.spint)

        start, apex, end = detect_peaks(roi.spint, roi.noise, roi.baseline)
        return [Peak(start=s, apex=a, end=e, roi=roi) for s, a, e in zip(start, apex, end)]



"""Base utilities for simulation."""

from __future__ import annotations

import random
from functools import cache, cached_property
from math import exp

import numpy
import pydantic

from ..chem import EM, Formula
from ..core.models import IsotopicEnvelope
from ..core.utils.numpy import FloatArray1D


class DataAcquisitionSpec(pydantic.BaseModel):
    """Define the acquisition parameters of a simulated."""

    grid: MZGridSpec | None = None
    """The m/z grid specification. If not specified, a grid is created using features m/z"""

    mz_std: pydantic.NonNegativeFloat = 0.0
    """Additive noise added to m/z in each scan"""

    int_std: pydantic.NonNegativeFloat = 0.0
    """additive noise added to spectral intensity on each scan"""

    mz_width: pydantic.PositiveFloat = 0.005
    """The peak width in the m/z domain. Used only when a grid specification is provided"""

    n_scans: pydantic.PositiveInt = 100
    """The number of scans in the sample"""

    time_resolution: pydantic.PositiveFloat = 1.0
    """The time spacing between scans"""

    min_int: pydantic.PositiveFloat | None = None
    """If specified, elements in a spectrum with intensity values lower than this parameter are removed"""

    ms_level: pydantic.PositiveInt = 1
    """The spectra MS level"""


class MZGridSpec(pydantic.BaseModel):
    """Define the minimum, maximum and spacing of m/z values in spectra."""

    low: pydantic.PositiveFloat = 100.0
    """The minimum m/z value in the grid"""

    high: pydantic.PositiveFloat = 1200.0
    """The maximum m/z value in the grid"""

    size: pydantic.PositiveInt = 10000
    """The number of elements in the grid"""

    def create(self) -> FloatArray1D:
        """Create a m/z grid."""
        return numpy.linspace(self.low, self.high, self.size)


class AbundanceSpec(pydantic.BaseModel):
    """Define the abundance of a chemical species in a sample.

    The abundance is computed as follows:

    1.  draw a value :math:`u` from the uniform distribution :math:`[0, 1]`.
    2.  If :math:`u` is lower than the `prevalence` field, set the abundance to zero.
    3.  Otherwise, set the abundance to a value sampled from a gaussian distribution.

    """

    mean: pydantic.PositiveFloat = 100.0
    """The mean of the gaussian distribution used to sample the abundance."""

    std: pydantic.PositiveFloat = 0.0
    """The standard deviation of the gaussian distribution used to sample the abundance"""

    prevalence: float = pydantic.Field(gt=0.0, le=1.0, default=1.0)
    """The probability of the species to be present in a sample."""

    def sample_abundance(self) -> float:
        """Get a realization of the abundance."""
        is_in_sample = random.uniform(0.0, 1.0) < self.prevalence
        c = max(0.0, random.gauss(mu=self.mean, sigma=self.std))  # force signal to be non-negative
        return c if is_in_sample else 0.0


class MeasurementNoiseSpec(pydantic.BaseModel):
    r"""Define an additive error term added to the measured signal of an adduct.

    The noise for a sample is computed as follows:

    1.  compute the :term:`snr` using the `base_snr` field and the isotopologue abundance as a
        scaling factor. Set the :term:`snr` to the maximum value between this value and the `min_snr`
        field.
    2.  Compute the noise level :math:`\sigma:` as the quotient between the signal and the :term:`snr`.
    3.  Sample the noise from a distribution :math:`~N(0, \sigma)`

    """

    base_snr: pydantic.PositiveFloat | None = None
    """The base :term:`snr` of the additive noise applied to isotopologue signal"""

    min_snr: pydantic.PositiveFloat = 10.0
    """The minimum :term:`snr` of the additive noise applied to isotopologues signal. This value allows
    to set a lower bound on the :term:`snr` for low intensity features. If the `snr` parameter is
    not set, this value is ignored."""

    @pydantic.model_validator(mode="after")
    def _ensure_snr_geq_min_snr(self):
        if self.base_snr is not None and self.base_snr < self.min_snr:
            raise ValueError("`snr` field must be greater or equal than `min_snr`.")
        return self

    def compute_snr(self, pk: float) -> float | None:
        """Compute the :term:`snr` level for an isotopologue signal.

        :param pk: the isotopologue relative abundance, used to scale the base :term:`snr`

        """
        return None if self.base_snr is None else max(self.min_snr, self.base_snr * pk)

    def sample_noise(self, signal: float, pk: float) -> float:
        """Draw a noise term sample for an isotopologue signal.

        :param signal: the observed signal, as the product of the instrument
            response, the base abundance and the isotopologue relative abundance.
        :param pk: the relative abundance of an isotopologue, used to scale the :term:`snr`
            for isotopologues with lower signals.

        """
        snr = self.compute_snr(pk)
        return 0.0 if snr is None else random.gauss(sigma=signal / snr)


class InstrumentResponseSpec(pydantic.BaseModel):
    r"""Define the instrument response factor for an adduct.

    Computed as the product of the base response factor, the inter-batch effect factor
    and the sensitivity loss factor.

    The default parameters of this specification will generate a response factor without
    sensitivity loss over time and no additive noise.

    Refer to the :ref:`simulation-guide` for more details.
    """

    model_config = pydantic.ConfigDict(frozen=True)

    base_response_factor: pydantic.PositiveFloat = 1.0
    """The adduct base response when no inter-batch or sensitivity loss effects are present."""

    max_sensitivity_loss: float = pydantic.Field(ge=0.0, le=1.0, default=0.0)
    """The maximum sensitive loss in an analytical batch."""

    sensitivity_decay: float = pydantic.Field(ge=0.0, default=0.0)
    """The decay parameter for the time-dependent sensitivity loss. We suggest to use values between
    ``0.001`` and ``1.0``, as larger values decays to the maximum sensitivity value too fast. This
    value should also be selected based on the batch size, as in longer batches it is possible
    to see the effect of smaller decay values.
    """

    interbatch_variation: float = pydantic.Field(ge=0.0, le=1.0, default=1.0)
    """A factor applied to all samples from the same analytical batch. The interbatch
    variation factor is random but equal for all observations within a batch. The value
    for a given batch is samples from a uniform distribution with minimum value equal to
    this parameter and maximum value equal to ``1.0``. In the default configuration the
    inter-batch factor is set to ``1.0`` always."""

    @cache
    def get_interbatch_factor(self, batch: int) -> float:
        return random.uniform(self.interbatch_variation, 1.0)

    def get_sensitivity_loss_factor(self, order: int) -> float:
        """Compute the sensitivity loss factor applied to the base response."""
        sensitivity_decay = self.max_sensitivity_loss * exp(-order * self.sensitivity_decay)
        return 1 - self.max_sensitivity_loss + sensitivity_decay

    def compute_response_factor(self, order: int, batch: int) -> float:
        """Compute the response factor for a specific sample run order and analytical batch.

        :param order: the relative run order within a batch.
        :param batch: the analytical batch number

        """
        interbatch_factor = self.get_interbatch_factor(batch)
        sensitivity_loss = self.get_sensitivity_loss_factor(order)
        return self.base_response_factor * sensitivity_loss * interbatch_factor


class BaseChemicalSpeciesSpec(pydantic.BaseModel):
    r"""Define the how the signal generated by a chemical species are computed.

    .. math::

        x_{j} = c * p_{j} * f + \epsilon

    Where :math:`x_{j}` is the signal intensity for the j-th isotopologue, :math:`c` is the
    abundance of the species that generates the adducts, :math:`p_{j}` is the abundance of the
    j-th isotopologue included in the simulation, :math:`f` is the response factor of the instrument
    and :math:`epsilon` is an additive error term.

    """

    formula: str
    """The ion formula. Used as a :py:class:`tidyms.chem.Formula` argument."""

    n_isotopologues: pydantic.PositiveInt = 1
    """The number of isotopologues to simulate."""

    abundance: dict[str, AbundanceSpec] | AbundanceSpec = AbundanceSpec()
    """Define the abundance :math:`c` of the chemical species that generates the ion. Multiple
    abundance specifications may be defined for different sample groups. In this case, the
    corresponding specification will be selected based on the simulated sample group. If the
    sample groups is not found a ``ValueError`` will be raised."""

    response: InstrumentResponseSpec = InstrumentResponseSpec()
    """Define how the response factor :math:`f` is computed"""

    noise: MeasurementNoiseSpec = MeasurementNoiseSpec()
    """Defines how the additive noise :math:`epsilon` is computed"""

    @pydantic.computed_field(repr=False)
    @cached_property
    def charge(self) -> int:
        return Formula(self.formula).charge

    @pydantic.computed_field(repr=False)
    @cached_property
    def _envelope(self) -> IsotopicEnvelope:
        return Formula(self.formula).get_isotopic_envelope(self.n_isotopologues)

    def compute_abundance(self, group: str | None = None) -> float:
        """Compute a realization of the species abundance in the specified group.

        :param group: the group name if multiple groups where provided for the abundance specification.
            If not provided a default group is chosen.
        :raises ValueError: if the group is not found in the specification

        """
        if group is None and isinstance(self.abundance, dict):
            try:
                group = list(self.abundance)[0]
            except IndexError as e:
                raise ValueError("Abundance spec cannot be an empty dictionary") from e

        if isinstance(self.abundance, dict):
            assert isinstance(group, str)
            try:
                return self.abundance[group].sample_abundance()
            except KeyError as e:
                raise ValueError(f"Group `{group}` is not defined in the abundance specification.") from e

        return self.abundance.sample_abundance()

    def get_mz(self) -> list[float]:
        """Compute the m/z of features in the adduct."""
        return [(Mk - self.charge * EM) / abs(self.charge) for Mk in self._envelope.mz]

    def compute_intensity(self, group: str | None = None, order: int = 0, batch: int = 0) -> list[float]:
        """Compute a realization of features intensity all isotopologues."""
        c = self.compute_abundance(group)
        intensity = list()
        for pk in self._envelope.p:
            f = self.response.compute_response_factor(order, batch)
            signal = c * f * pk
            noise = self.noise.sample_noise(signal, pk)
            x = max(0.0, signal + noise)  # force intensity to be non-negative
            intensity.append(x)
        return intensity



"""Utilities to validate and score isotopic envelopes."""

from __future__ import annotations

from typing import Callable, Sequence

import numpy as np
from scipy.special import erfc

from ..core.utils.numpy import FloatArray1D
from .config import EnvelopeScorerConfiguration, EnvelopeValidatorConfiguration
from .context import DEFAULT_CONTEXT, ChemicalContext
from .envelope_utils import (
    Envelope,
    EnvelopeQuery,
    EnvelopeQueryTolerance,
    HomoAtomicEnvelopeCache,
    combine_envelopes,
)
from .formula_generator import FormulaCoefficients, FormulaGenerator
from .table import PeriodicTable


class _EnvelopeGenerator:
    """Base class to generate envelopes from a list of molecular formulas."""

    def __init__(self, config: EnvelopeScorerConfiguration):
        if config.context is None:
            self._context = DEFAULT_CONTEXT
        else:
            self._context = ChemicalContext(config.context)

        self.config = config

        self._formula_generator = FormulaGenerator(config)
        self._pos_env = make_formula_coefficients_envelopes(
            config.bounds,
            self._formula_generator.pos,
            self._context.envelope_cache,
            self._context.table,
            config.max_length,
        )

        self._neg_env = make_formula_coefficients_envelopes(
            config.bounds,
            self._formula_generator.neg,
            self._context.envelope_cache,
            self._context.table,
            config.max_length,
        )
        self._c_env = _create_12c_envelope(
            self._formula_generator, self._context.table, self._context.envelope_cache, config.max_length
        )

        self._query: EnvelopeQuery | None = None
        self.results: Envelope | None = None

    def generate_envelopes(self, M: Sequence[float], p: Sequence[float], tolerance: float):
        """Compute isotopic envelopes for formula candidates generated using the MMI mass from the envelope.

        :param M: the exact mass of the envelope
        :param p: the envelope normalized abundances.
        :param tolerance: mass tolerance to generate formulas.

        """
        if len(M) > self.config.max_length:
            msg = "`max_length` ({}) is lower than the query length ({})"
            raise ValueError(msg)

        query = EnvelopeQuery(M=M[: self.config.max_length], p=p[: self.config.max_length])
        self._formula_generator.generate_formulas(query.get_mmi_mass(), tolerance)
        query_envelopes = _find_result_envelopes(self._formula_generator, self._pos_env, self._neg_env, self._c_env)
        self._query = query
        self.results = query_envelopes

    def filter(self, min_M_tol: float, max_M_tol: float, p_tol: float, k: int):
        r"""Filter values from the k-th envelope that are outside the specified bounds.

        :param min_M_tol: mass values lower than this value are filtered.
        :param max_M_tol: mass values greater than this value are filtered.
        :param p_tol: abundance tolerance
        :param k: the envelope index to filter values.

        Notes
        -----
        Envelopes are filtered based on the following inequality. For each i-th
        peak the m/z tolerance is defined as follows:

        .. math::

            t_{i} = t + (T - t)(1 - y_{i})

        where :math:`t_{i}` is the mass tolerance for the i-th peak, t is the
        `min_mz_tolerance`, T is the `max_mz_tolerance` and :math:`y_[i]` is
        the abundance of the i-th value. Using this tolerance, an interval is
        built for the query mass, and candidates outside this interval are
        removed. This approach accounts for greater m/z errors for lower
        intensity peaks in the envelope.

        """
        if self.results is None or self._query is None:
            msg = "Envelopes must be generated first using the generate method."
            raise ValueError(msg)
        tol = EnvelopeQueryTolerance(min_M_tol=min_M_tol, max_M_tol=max_M_tol, p_tol=p_tol)
        bounds = self._query.get_bounds(tol, k)
        self.results.exclude(bounds)


class EnvelopeValidator(_EnvelopeGenerator):
    r"""Envelope validator.

    Notes
    -----
    Envelope validation is performed as follows:

    1.  For a query envelope mass and abundance `Mq`and `pq`, all formulas
        compatibles with the MMI are computed (see FormulaGenerator).
    2.  For each i-th pair of `Mq` and `pq`, a mass tolerance and abundance
        tolerance is defined as follows:

        .. math::
            dM_{i} = dM^{\textrm{max}} * pq_{i} + dM^{\textrm{min}} (1 - pq_{i})

        Where :math:`dM^{\textrm{max}}` is `min_M_tol`, :math:`dM^{\textrm{min}}`
        is `max_M_tol` and :math:`pq_{i}` is the i-th query abundance.
        Using the mass tolerance and abundance tolerance, candidates with
        mass or abundance values outside this interval are removed.
    3.  The candidates that remains define a mass and abundance window for
        the i + 1 elements of `Mq` and `pq`. If the values fall inside the
        window, the i + 1 elements are validated and the procedure is repeated
        until all isotopologues are validated or until an invalid isotopologue
        is found.

    """

    def __init__(self, config: EnvelopeValidatorConfiguration):
        config = config.model_copy(deep=True)
        config.remove_elements_with_a_single_isotope()
        super(EnvelopeValidator, self).__init__(config)
        self.config = config

    def validate(self, M: Sequence[float], p: Sequence[float]) -> int:
        """Validate an envelope."""
        valid_length = 0
        tol = p[0] * self.config.min_M_tol + (1 - p[0]) * self.config.max_M_tol
        self.generate_envelopes(M, p, tol)
        if self.results is None:
            return 0
        query = self._query
        assert query is not None
        length = len(query.M)

        while (length >= 2) and (valid_length <= 1):
            for k in range(length):
                self.filter(self.config.min_M_tol, self.config.max_M_tol, self.config.p_tol, k)
                if self.results.get_n_rows():
                    valid_length = k + 1
                else:
                    break
            if (valid_length <= 1) and (length > 2):
                valid_length = 0
                query.crop()
                length -= 1
                self.results.reset_exclude()
                self.results.crop(length)
            else:
                break

        return valid_length


class EnvelopeScorer(_EnvelopeGenerator):
    r"""Rank formula candidates by comparing measured and theoretical isotopic envelopes.

    Refer to the :ref:`user guide <scoring-formulas-guide>` for details usage instructions.

    :param config: the envelope generator configuration
    :param scorer: function that scores formula candidate envelopes. If ``None``, the
        function :func:`score_envelope` is used. A custom scoring function can be passed
        with the following signature:

        .. code-block:: python

            def score(M, p, Mq, pq, **kwargs):
                pass

        where `M` and `p` are arrays of the formula candidates exact mass and abundances and
        `Mq` and `pq` are the query mass and query abundance.

    kwargs :
        Optional parameter to pass into the scoring function.

    """

    def __init__(self, config: EnvelopeScorerConfiguration, scorer: Callable | None = None, **kwargs):
        super(EnvelopeScorer, self).__init__(config)

        if callable(scorer):
            self.scorer = scorer
            self.scorer_params = kwargs
        else:
            self.scorer = score_envelope
            self.scorer_params = kwargs
        self.scores: FloatArray1D | None = None

    def score(self, M: Sequence[float], p: Sequence[float], tol: float):
        """Score the isotopic envelope.

        The results can be recovered using the `get_top_results` method.

        Formulas are generated assuming that the first element in the envelope
        is the minimum mass isotopologue.

        :param M: exact mass of the envelope.
        :param p: abundance of the envelope.
        :param tol: mass tolerance used in formula generation.

        """
        self.generate_envelopes(M, p, tol)
        assert self.results is not None
        assert self._query is not None
        n_results = self.results.get_n_rows()
        scores = np.zeros(n_results)

        query_M = np.array(self._query.M)
        query_p = np.array(self._query.p)

        for i, (Mi, pi) in enumerate(self.results.iterate_rows()):
            scores[i] = self.scorer(Mi, pi, query_M, query_p, **self.scorer_params)
        self.scores = scores

    def get_top_results(self, n: int | None = 10):
        """Fetch the top ranked formula candidates and their score.

        :param n: number of first n results to return. If ``None``, return all formula candidates.

        """
        coefficients, elements, _ = self._formula_generator.results_to_array()

        # sort coefficients using the score and keep the first n values
        assert self.scores is not None
        top_n_index = np.argsort(self.scores)
        if n is not None:
            top_n_index = top_n_index[: (-n - 1) : -1]

        scores = self.scores[top_n_index]
        coefficients = coefficients[top_n_index]
        return coefficients, elements, scores


def score_envelope(
    M: np.ndarray,
    p: np.ndarray,
    Mq: np.ndarray,
    pq: np.ndarray,
    min_sigma_M: float = 0.01,
    max_sigma_M: float = 0.01,
    min_sigma_p: float = 0.05,
    max_sigma_p: float = 0.05,
):
    r"""Score the similarity between two isotopes.

    :param M: theoretical mass values.
    :param p: theoretical abundances.
    :param Mq:  query Mass values
    :param pq: query abundances.
    :param min_sigma_M: minimum mass standard deviation
    :param max_sigma_M: maximum mass standard deviation
    :param min_sigma_p: minimum abundance standard deviation.
    :param max_sigma_p: maximum abundance standard deviation.

    Returns
    -------
    score : float
        Number between 0 and 1. Higher values are related with similar envelopes.

    Notes
    -----
    The query envelope is compared against the theoretical envelope assuming
    a likelihood approach, similar to the described in [1]_. It is assumed
    that the theoretical mass and abundance is a normal random variable,
    with mean values defined by `M` and `p` and standard deviation computed as
    follows:

    .. math::

        \sigma_{M,i} = p_{i} \sigma_{M}^{\textrm{max}} + (1 - p_{i}) \sigma_{M}^{\textrm{min}}

    Where :math:`\sigma_{M,i}` is the standard deviation for the i-th element
    of `M`, :math:`p_{i}` is the i-th element of `p`, :math:`\sigma_{M}^{\textrm{max}}`
    is `max_sigma_M` and :math:`\sigma_{M}^{\textrm{min}}` is `min_sigma_M`. An
    analogous computation is done to compute the standard deviation for each
    abundance. Using this values, the likelihood of generating the values `Mq` and
    `pq` from `M` and `p` is computed using the error function.

    References
    ----------
    ..  [1] Sebastian Böcker, Matthias C. Letzel, Zsuzsanna Lipták, Anton
        Pervukhin, SIRIUS: decomposing isotope patterns for metabolite
        identification, Bioinformatics, Volume 25, Issue 2, 15 January 2009,
        Pages 218–224, https://doi.org/10.1093/bioinformatics/btn603

    """
    mz_sigma = max_sigma_M + (min_sigma_M - max_sigma_M) * pq
    sp_sigma = max_sigma_p + (min_sigma_p - max_sigma_p) * pq
    M = M[: Mq.size]
    p = p[: Mq.size]
    # normalize again the candidate intensity to 1
    p = p / p.sum()

    # corrects overestimation of the first peak area. This is done computing
    # an offset factor to subtract to the first peak. This correction is applied
    # only if the offset is positive. The offset value is computed in a way to
    # satisfy two conditions: the abundance of the first peak is equal to the
    # abundance of the candidate peak and the total area is normalized to one.
    # offset = (spq[0] - sp[0]) / (1 - sp[0])
    # offset = max(0, offset)
    norm = (pq[0] - 1) / (p[0] - 1)
    # spq = spq / (1 - offset)
    if norm < 1:
        pq = pq / norm
        pq[0] = p[0]

    # add a max offset parameter

    Mq = Mq + M[0] - Mq[0]
    dmz = np.abs(M - Mq) / (np.sqrt(2) * mz_sigma)
    dmz = dmz[pq > 0]
    dsp = np.abs(p - pq) / (np.sqrt(2) * sp_sigma)
    score = erfc(dmz).prod() * erfc(dsp).prod()
    return score


def make_formula_coefficients_envelopes(
    bounds: dict[str, tuple[int, int]],
    coefficients: FormulaCoefficients,
    cache: HomoAtomicEnvelopeCache,
    table: PeriodicTable,
    max_length: int,
) -> Envelope:
    """Compute the isotopic envelopes for coefficient formulas."""
    # initialize envelopes
    rows = coefficients.coefficients.shape[0]
    M_arr = np.zeros((rows, max_length))
    p_arr = np.zeros((rows, max_length))
    p_arr[:, 0] = 1
    env = Envelope(M=M_arr, p=p_arr)

    for k, isotope in enumerate(coefficients.isotopes):
        # if a symbol is passed in bounds, e.g. "C", the column in coefficients
        # will be the isotope "12C". to find the abundance, the element symbol
        # is used in bounds. If an isotope is specified, it is assumed that it
        # has no envelope.
        if isotope.to_str() in bounds:
            lb, ub = bounds[isotope.to_str()]
            k_env = cache.make_envelope_array(isotope, lb, ub, max_length)
        elif isotope.symbol in bounds:
            lb, ub = bounds[isotope.symbol]
            element = table.get_element(isotope.z)
            k_env = cache.make_envelope_array(element, lb, ub, max_length)
        else:
            # ignore dummy elements used to solve the formula generation problem
            # This occurs only in cases where there are only isotopes with positive
            # or negative mass defects.
            continue
        # Create copies of each row based on the values in the coefficient arrays
        # using lb correct index values in cases when 0 is not the lower bound
        k_env.M = k_env.M[coefficients.coefficients[:, k] - lb, :]
        k_env.p = k_env.p[coefficients.coefficients[:, k] - lb, :]
        env = combine_envelopes(env, k_env)
    return env


def _find_result_envelopes(
    fg: FormulaGenerator,
    pos_env: Envelope,
    neg_env: Envelope,
    c_env: Envelope,
) -> Envelope | None:
    length = pos_env.M.shape[1]
    shape = (fg.get_n_results(), length)
    M = np.zeros(shape, dtype=float)
    p = np.zeros(shape, dtype=float)
    start = 0
    for kp_index, kn_index, kc_index in fg.get_results().values():
        k_size = kp_index.size
        if k_size > 0:
            kp_env = Envelope(M=pos_env.M[kp_index], p=pos_env.p[kp_index])
            kn_env = Envelope(M=neg_env.M[kn_index], p=neg_env.p[kn_index])
            kc_env = Envelope(M=c_env.M[kc_index], p=c_env.p[kc_index])

            k_env = combine_envelopes(kp_env, kn_env)
            k_env = combine_envelopes(k_env, kc_env)

            end = start + k_size
            M[start:end] = k_env.M
            p[start:end] = k_env.p
            start = end

    envelopes = Envelope(M=M, p=p)
    if M.size < 2:
        return None
    envelopes.crop(M.size)
    return envelopes


def _create_12c_envelope(
    formula_generator: FormulaGenerator, table: PeriodicTable, cache: HomoAtomicEnvelopeCache, length: int
) -> Envelope:
    c12 = table.get_isotope("12C")
    c = table.get_element("C")
    c12_bounds = formula_generator.bounds.bounds.get(c12)
    if c12_bounds is None:
        nc_min, nc_max = 0, 0
    else:
        nc_min, nc_max = c12_bounds.lower, c12_bounds.upper
    return cache.make_envelope_array(c, nc_min, nc_max, length)



import numpy as np
import matplotlib.pyplot as plt

from tidyms2.core.utils.numpy import gauss, gaussian_mixture
from tidyms2.algorithms.signal import estimate_baseline, estimate_noise, detect_peaks, smooth

np.random.seed(1234)
signal_height = 100
snr = 10
n_col = 4
x = np.arange(200)
noise_level = signal_height / snr
noise = np.random.normal(size=x.size, scale=noise_level)
fig, ax = plt.subplots(
    nrows=3, ncols=n_col, figsize=(12, 12), sharex=True, sharey=True)

# first row: one peak, different baselines
row = 0
baselines = [4, gauss(x, 100, 40, 20), x ** 2 * 0.002, np.sin(x * np.pi / 400) * 50]
for col in range(n_col):
    signal = gauss(x, 100, 3, signal_height)
    y = signal + noise
    noise_estimation = estimate_noise(y)
    ys = smooth(y, 1)
    baseline = estimate_baseline(ys, noise_estimation)
    start, _, end = detect_peaks(ys, noise_estimation, baseline)
    ax[row, col].plot(x, y)
    ax[row, col].plot(x, baseline)
    for s, e in zip(start, end):
        ax[row, col].fill_between(x[s:e + 1], baseline[s:e + 1], y[s:e + 1], alpha=0.25)

# second row: two peaks, same baselines as first row
row = 1
for col in range(n_col):
    signal = gaussian_mixture(x, (100, 3, signal_height), (110, 3, signal_height))
    y = signal + baselines[col] + noise
    noise_estimation = estimate_noise(y)
    ys = smooth(y, 1)
    baseline = estimate_baseline(ys, noise_estimation)
    start, apex, end = detect_peaks(ys, noise_estimation, baseline)
    ax[row, col].plot(x, y)
    ax[row, col].plot(x, baseline)
    for s, e in zip(start, end):
        ax[row, col].fill_between(x[s:e + 1], baseline[s:e + 1], y[s:e + 1], alpha=0.25)

# third row: different peak widths:
row = 2
widths = [3, 5, 7, 10]
for col in range(n_col):
    w = widths[col]
    signal = gauss(x, 100, w, signal_height)
    y = signal + baselines[0] + noise
    noise_estimation = estimate_noise(y)
    ys = smooth(y, 1)
    baseline = estimate_baseline(ys, noise_estimation)
    start, apex, end = detect_peaks(ys, noise_estimation, baseline)
    ax[row, col].plot(x, y)
    ax[row, col].plot(x, baseline)
    for s, e in zip(start, end):
        ax[row, col].fill_between(x[s:e + 1], baseline[s:e + 1], y[s:e + 1], alpha=0.25)



"""TidyMS constants."""

import enum


class OperatorType(str, enum.Enum):
    """Available operators types."""

    SAMPLE = "sample"
    """Apply operations on sample storage."""

    ASSAY = "assay"
    """Apply operations on assay storage."""

    MATRIX = "matrix"
    """Apply operations on data matrix."""


class SeparationMode(str, enum.Enum):
    """Analytical method separation platform."""

    DART = "dart"
    """direct analysis in real time ion source"""

    HPLC = "hplc"
    """High Performance Liquid Chromatography separation"""

    UPLC = "uplc"
    """ultra performance liquid chromatography separation"""


class MSInstrument(enum.Enum):
    """Available MS instrument types."""

    QTOF = "qtof"
    ORBITRAP = "orbitrap"


class Polarity(str, enum.Enum):
    """Scan polarity."""

    POSITIVE = "positive"
    NEGATIVE = "negative"


class MSDataMode(str, enum.Enum):
    """Raw data mode."""

    PROFILE = "profile"
    CENTROID = "centroid"


class SampleType(str, enum.Enum):
    """Sample types in an untargeted metabolomics assay."""

    SAMPLE = "sample"
    """A test subject sample."""

    TECHNICAL_QC = "qc"
    """A technical QC sample"""

    EXPERIMENTAL_QC = "eqc"
    """An experimental QC sample"""

    DILUTED_QC = "dqc"
    """A diluted QC sample"""

    BLANK = "blank"
    """A blank QC sample"""

    OTHER = "other"
    """Samples that do not belong to any of the other categories."""


class ScalingMethod(str, enum.Enum):
    """Available feature scaling methods.

    All scaling methods support matrix with NaN values. NaN values will be obtained when scaling columns
    with constant values.

    """

    AUTOSCALING = "autoscaling"
    r"""Scale features to unitary population variance:

    .. math:

            X_{j}^{\textrm{(scaled)}} = \frac{X_{j} - \bar{X}_{j}}{S_{j}}

    where :math:`X_{j}` is the j-th column of the matrix, :math:`\bar{X}_{j}` is the column average and
    :math:`S_{j}` is the column population standard deviation.

    """

    PARETO = "pareto"
    r"""Mean center and scale features using the square root of the population standard deviation:

    .. math:

            X_{j}^{\textrm{(scaled)}} = \frac{X_{j} - \bar{X}_{j}}{\sqrt{S_{j}}}

    where :math:`X_{j}` is the j-th column of the matrix, :math:`\bar{X}_{j}` is the column average and
    :math:`S_{j}` is the column population standard deviation.

    """

    RESCALING = "rescaling"
    r"""Scale features to the range :math:`[0, 1]`:

    .. math:

            X_{j}^{\textrm{(scaled)}} = \frac{X_{j} - \min{X}_{j}}{\max{X_{j}} - \min{X_{j}}}

    where :math:`X_{j}` is the j-th column of the matrix, :math:`\max{X}_{j}` is the column maximum and
    :math:`\min{X}_{j}` is the column minimum.

    """


class AggregationMethod(str, enum.Enum):
    """Available sample aggregations methods."""

    LOD = "lod"
    """Compute the limit of detection, defined as the mean plus three times the sample standard
    deviation."""

    LOQ = "loq"
    """Compute the limit of quantification, defined as the mean plus ten times the sample standard
    deviation."""

    SUM = "sum"
    """sums the feature abundance from samples"""

    MEAN = "avg"
    """Computes the mean feature abundance from samples"""

    MEDIAN = "median"
    """Computes the median feature abundance across samples"""

    MIN = "min"
    """computes the minimum features value across samples"""

    MAX = "max"
    """computes the maximum features value across samples"""

    STD = "std"
    """compute the standard deviation of features across samples"""


class NormalizationMethod(str, enum.Enum):
    """Available sample normalization methods."""

    SUM = "sum"
    """Normalize samples using sum of all features."""

    MAX = "max"
    """Normalize samples using the maximum value of all features"""

    EUCLIDEAN = "euclidean"
    """Normalize samples using the row 2-norm."""

    FEATURE = "feature"
    """Normalize samples using the value of a feature."""


class CorrelationMethod(str, enum.Enum):
    """Available correlation methods."""

    PEARSON = "pearson"
    """The Pearson's correlation coefficient."""

    SPEARMAN = "spearman"
    """The Spearman rank coefficient."""



"""Tools for working with chemical formulas."""

from __future__ import annotations

import string
from collections import Counter
from copy import copy
from typing import overload

from ..core.models import IsotopicEnvelope
from .atoms import EM, Isotope
from .context import DEFAULT_CONTEXT, ChemicalContext
from .envelope_utils import Envelope, HomoAtomicEnvelopeCache, combine_envelopes
from .exceptions import InvalidFormula, IsotopeError
from .table import PeriodicTable


class Formula:
    """Represent a chemical formula as a mapping from isotopes to formula coefficients.

    Refer to the :ref:`user guide <working-with-formulas-guide>` for usage instructions.

    :param formula: a string representation of a formula or a mapping from isotopes to formula coefficients
    :param charge: the numerical charge of the formula
    :param context: the chemical context where data is fetch from. In the majority of cases this parameter
        may be ignored. This parameter should be set only if working with custom isotopic abundances.

    >>> Formula("H2O")
    Formula(H2O)
    >>> Formula("(13C)O2")
    Formula((13C)O2)
    >>> Formula("[Cr(H2O)6]3+")
    Formula([H12CrO6]3+)
    >>> Formula("CH3CH2CH3")
    Formula(C3H8)
    >>> Formula({"C": 1, "17O": 2})
    Formula(C(17O)2)

    """

    @overload
    def __init__(self, formula: dict[Isotope, int], charge: int | None, context: None = None): ...

    @overload
    def __init__(self, formula: dict[str, int], charge: int | None, context: ChemicalContext | None = None): ...

    @overload
    def __init__(self, formula: str, charge: None = None, context: ChemicalContext | None = None): ...

    def __init__(self, formula, charge=None, context: ChemicalContext | None = None):
        self.context = DEFAULT_CONTEXT if context is None else context
        if isinstance(formula, str):
            formula_str, charge = _parse_charge(formula)
            composition = _parse_formula(formula_str, self.context.table)
        else:
            f_composition = formula
            charge = 0 if charge is None else charge
            composition: Counter[Isotope] = Counter()
            for k, v in f_composition.items():
                if isinstance(k, str):
                    isotope = self.context.table.get_isotope(k)
                elif isinstance(k, Isotope):
                    isotope = k
                else:
                    msg = "Composition keys must be a string representation of an isotope or an isotope object"
                    raise IsotopeError(msg)

                if not isinstance(v, int) or (v < 1):
                    msg = "Formula coefficients must be positive integers"
                    raise ValueError(msg)
                composition[isotope] = v

            if not isinstance(charge, int):
                msg = "Charge must be an integer"
                raise ValueError(msg)
        self.charge = charge
        self.composition = composition

    def __add__(self, other: Formula) -> Formula:
        sum_composition = copy(self.composition)
        sum_composition.update(other.composition)
        sum_charge = self.charge + other.charge
        return Formula(sum_composition, sum_charge)

    def __sub__(self, other: Formula) -> Formula:
        comp = copy(self.composition)
        comp.subtract(other.composition)
        charge = self.charge - other.charge
        min_coeff = min(comp.values())
        if min_coeff < 0:
            msg = "subtraction cannot generate negative coefficients"
            raise ValueError(msg)
        comp = Counter({k: v for k, v in comp.items() if v > 0})
        diff_f = Formula(comp, charge)
        return diff_f

    def __eq__(self, other) -> bool:
        return (self.charge == other.charge) and (self.composition == other.composition)

    def get_exact_mass(self) -> float:
        """Compute the exact mass of the formula.

        Returns
        -------
        exact_mass: float

        Examples
        --------
        >>> import tidyms as ms
        >>> f = ms.chem.Formula("H2O")
        >>> f.get_exact_mass()
        18.010564684

        """
        exact_mass = sum(x.m * k for x, k in self.composition.items())
        exact_mass -= EM * self.charge
        return exact_mass

    def get_nominal_mass(self) -> int:
        """Compute the nominal mass of the formula.

        >>> import tidyms as ms
        >>> f = ms.chem.Formula("H2O")
        >>> f.get_nominal_mass()
        18

        """
        nominal_mass = sum(x.a * k for x, k in self.composition.items())
        return nominal_mass

    def get_isotopic_envelope(self, n: int = 10, min_p: float = 1e-10) -> IsotopicEnvelope:
        """Compute the isotopic envelope of the formula.

        The natural abundance is assumed for each monoisotope, i.e., the most
        abundant isotope. If others isotopes are present in the formula, they
        are assumed to have abundance equal to 1. See the examples for a
        clarification of this.

        :param n: the number of isotopes to include in the results.
        :param min_p: isotopes are included until the abundance is lower than this value

        If no isotopes are specified in the formula, the natural abundance is used.

        >>> import tidyms as ms
        >>> f = ms.chem.Formula("C6H6")
        >>> print(f)
        C6H6
        >>> f.get_isotopic_envelope(n=3)
        (array([78.04695019, 79.05033578, 80.05373322]),
         array([0.93686877, 0.06144402, 0.00168606]))

        Using isotopes other than the monoisotope are treated as if they have
        an abundance equal to one.

        >>> f = ms.chem.Formula("(13C)6(2H)6")
        >>> print(f)
        (13C)6(2H)6
        >>> f.get_isotopic_envelope()
        (array([90.10473971]), array([1.]))

        """
        envelope = _find_formula_envelope(
            self.composition, self.context.envelope_cache, self.context.table, n, min_p=min_p
        )
        M = envelope.M - EM * self.charge
        return IsotopicEnvelope(mz=M.tolist(), p=envelope.p.tolist())  # type: ignore

    def __repr__(self):
        return "Formula({})".format(str(self))

    def __str__(self):
        return _get_formula_str(self.composition, self.charge, self.context.table)


_matching_parenthesis = {"(": ")", "[": "]"}


def _multiply_formula_coefficients(composition: Counter, multiplier: int):
    if multiplier != 1:
        for k in composition:
            composition[k] *= multiplier


def _parse_charge(formula: str) -> tuple[str, int]:
    """Compute the charge state of a formula and remove the charge from the formula string.

    :param formula: molecular formula string

    Returns
    -------
    formula_without_charge, charge: str, int

    """
    # check if there's a charge in the formula
    if formula[-1] == "+":
        sign = 1
    elif formula[-1] == "-":
        sign = -1
    else:
        sign = 0

    # for charge with an absolute value greater than 1 enforces the use
    # of parenthesis to prevent ambiguity with a formula coefficient
    if sign and (formula[-2] in string.digits):
        try:
            matching = _matching_parenthesis[formula[0]]
            start = 1
            end = formula.rfind(matching)
            charge = sign * int(formula[end + 1 : -1])
            return formula[start:end], charge
        except (KeyError, ValueError):
            raise InvalidFormula
    elif sign:
        return formula[:-1], sign
    else:
        return formula, 0


def _get_token_type(formula: str, ind: int) -> int:
    """Assign 0 to elements, 1 to isotopes and 2 to expressions.

    Return the token type and a matching parenthesis if necessary
    """
    c = formula[ind]
    if c in string.ascii_uppercase:
        token_type = 0
    elif c in "(":
        if formula[ind + 1] in string.digits:
            token_type = 1
        else:
            token_type = 2
    elif c == "[":
        token_type = 2
    else:
        raise InvalidFormula
    return token_type


def _find_matching_parenthesis(formula: str, ind: int):
    parenthesis_open = formula[ind]
    parenthesis_close = _matching_parenthesis[parenthesis_open]
    match_ind = ind + 1
    level = 1
    try:
        while level > 0:
            c = formula[match_ind]
            if c == parenthesis_open:
                level += 1
            elif c == parenthesis_close:
                level -= 1
            match_ind += 1
        return match_ind - 1
    except IndexError:
        msg = "Formula string has non-matching parenthesis"
        raise InvalidFormula(msg)


def _get_coefficient(formula: str, ind: int):
    """Traverses a formula string to compute a coefficient.

    `ind` is a position after an element or expression.

    coefficient : int
    new_ind : int, new index to continue parsing the formula
    """
    length = len(formula)
    if (ind >= length) or (formula[ind] not in string.digits):
        coefficient = 1
        new_ind = ind
    else:
        end = ind + 1
        while (end < length) and (formula[end] in string.digits):
            end += 1
        coefficient = int(formula[ind:end])
        new_ind = end
    return coefficient, new_ind


def _tokenize_element(formula: str, ind: int, table: PeriodicTable):
    length = len(formula)
    if (ind < length - 1) and (formula[ind + 1] in string.ascii_lowercase):
        end = ind + 2
    else:
        end = ind + 1
    symbol = formula[ind:end]
    isotope = table.get_isotope(symbol)
    coefficient, end = _get_coefficient(formula, end)
    token = {isotope: coefficient}
    return token, end


def _tokenize_isotope(formula: str, ind: int, table: PeriodicTable):
    """Convert an isotope substring starting at `ind` index into a token.

    Returns
    -------
    token, new_ind

    """
    end = _find_matching_parenthesis(formula, ind)
    isotope = table.get_isotope(formula[ind + 1 : end])
    coefficient, end = _get_coefficient(formula, end + 1)
    token = {isotope: coefficient}
    return token, end


def _parse_formula(formula: str, table: PeriodicTable) -> Counter[Isotope]:
    """Parse a formula string into a Counter that maps isotopes to formula coefficients."""
    ind = 0
    n = len(formula)
    composition = Counter()
    while ind < n:
        token_type = _get_token_type(formula, ind)
        if token_type == 0:
            token, ind = _tokenize_element(formula, ind, table)
        elif token_type == 1:
            token, ind = _tokenize_isotope(formula, ind, table)
        else:
            # expression type evaluated recursively
            exp_end = _find_matching_parenthesis(formula, ind)
            token = _parse_formula(formula[ind + 1 : exp_end], table)
            exp_coefficient, ind = _get_coefficient(formula, exp_end + 1)
            _multiply_formula_coefficients(token, exp_coefficient)
        composition.update(token)
    return composition


# functions to get a formula string from a Formula


def _arg_sort_elements(symbol_list: list[str], mass_number_list: list[int]):
    """Return the sorted index for a list of elements symbols and mass numbers.

    If there are repeated elements, they are sorted by mass number.
    """
    zipped = list(zip(symbol_list, mass_number_list))
    return sorted(range(len(symbol_list)), key=lambda x: zipped[x])


def _get_formula_str(composition: Counter[Isotope], charge: int, table: PeriodicTable):
    # get C str
    c12 = table.get_isotope("12C")
    c13 = table.get_isotope("13C")
    h1 = table.get_isotope("1H")
    h2 = table.get_isotope("2H")
    ch = [c12, c13, h1, h2]

    # ADD C and H to formula str
    f_str = ""
    for i in ch:
        if i in composition:
            f_str += _isotope_coeff_to_f_str(i, composition[i], table.is_monoisotope(i))

    # add other elements, sorted alphabetically
    isotopes = set(composition)
    isotopes = isotopes.difference(ch)
    for i in sorted(isotopes, key=lambda x: x.symbol + str(x.a)):
        f_str += _isotope_coeff_to_f_str(i, composition[i], table.is_monoisotope(i))

    if charge:
        charge_str = _get_charge_str(charge)
        f_str = "[{}]{}".format(f_str, charge_str)

    return f_str


def _isotope_coeff_to_f_str(isotope: Isotope, coeff: int, is_monoisotope: bool) -> str:
    coeff_str = str(coeff) if coeff > 1 else ""
    symbol_str = isotope.symbol if is_monoisotope else f"({isotope.a}{isotope.symbol})"
    return f"{symbol_str}{coeff_str}"


def _get_charge_str(q: int) -> str:
    qa = abs(q)
    q_sign = "+" if q > 0 else "-"
    q_str = str(qa) if qa > 1 else ""
    q_str = q_str + q_sign
    return q_str


def _find_formula_envelope(
    composition: Counter[Isotope],
    cache: HomoAtomicEnvelopeCache,
    table: PeriodicTable,
    length: int,
    min_p: float = 1e-10,
) -> Envelope:
    """Compute the isotopic envelope for a formula."""
    # initialize an empty envelope for the formula
    env = cache.get_empty(length)

    for isotope, coeff in composition.items():
        if table.is_monoisotope(isotope):
            element = table.get_element(isotope.z)
            isotope_envelope = cache.get_envelope(element, coeff)
        else:
            isotope_envelope = cache.get_envelope(isotope, coeff)
        env = combine_envelopes(env, isotope_envelope)

    min_p_filter = env.p >= min_p
    env.p = env.p[min_p_filter].flatten()
    norm = env.p.sum()
    if norm > 0.0:
        env.p /= norm
    env.M = env.M[min_p_filter].flatten()
    return env



import matplotlib.pyplot as plt

from tidyms2.simulation.lcms import SimulatedLCMSSampleFactory
from tidyms2.io import MSData

adducts_spec = [
    {
        "formula": "[C54H104O6]+",
        "rt": {"mean": 10.0},
    },
    {
        "formula": "[C27H40O2]+",
        "rt": {"mean": 20.0},
    },
    {
        "formula": "[C24H26O12]+",
        "rt": {"mean": 30.0},
    },
]

simulated_sample_factory = SimulatedLCMSSampleFactory(adducts=adducts_spec) # type: ignore
sample = simulated_sample_factory(id="my_sample")
ms_data = MSData(sample)

sp = ms_data.get_spectrum(10)

fig, ax = plt.subplots()
ax.stem(sp.mz, sp.int, basefmt=" ", markerfmt="")
ax.set_xlabel("m/z")
ax.set_ylabel("Intensity [au]")
ax.set_xlim(845, 855)
ax.set_ylim(0, 50)



import matplotlib.pyplot as plt

from tidyms2.simulation.lcms import SimulatedLCMSSampleFactory
from tidyms2.io import MSData
from tidyms2.algorithms.raw import make_chromatograms, MakeChromatogramParameters

factory_spec = {
    "data_acquisition": {
        "n_scans": 100,
        "time_resolution": 0.5,
        "int_std": 2.0,
    },
    "adducts": [
        {
            "formula": "[C54H104O6]+",
            "n_isotopologues": 3,
            "abundance": {
                "mean": 500,
                "std": 50,
            },
            "rt": {
                "mean": 10.0,
                "std": 1.0,
                "width": 4.0,
            },
            "noise": {
                "snr": 200,
            }
        },
        {
            "formula": "[C27H40O2]+",
            "n_isotopologues": 2,
            "abundance": {
                "mean": 200,
                "std": 50,
                "prevalence": 0.7,
            },
            "rt": {"mean": 20.0},
        },
        {
            "formula": "[C24H26O12]+",
            "n_isotopologues": 5,
            "rt": {"mean": 30.0},
        },
    ],
}

simulated_sample_factory = SimulatedLCMSSampleFactory(**factory_spec)
sample = simulated_sample_factory(id="my_sample")
ms_data = MSData(sample)

mz = ms_data.get_spectrum(10).mz[0] # get the m/z value from the first feature


params = MakeChromatogramParameters(mz=[mz])
chrom = make_chromatograms(ms_data, params)[0]

fig, ax = plt.subplots()

ax.plot(chrom.time, chrom.int)
ax.set_xlabel("Rt [s]")
ax.set_ylabel("Intensity [au]")



import matplotlib.pyplot as plt

from tidyms2.simulation.lcms import SimulatedLCMSSampleFactory
from tidyms2.io import MSData
from tidyms2.algorithms.raw import make_chromatograms, MakeChromatogramParameters

adducts_spec = [
    {
        "formula": "[C54H104O6]+",
        "rt": {"mean": 10.0},
    },
    {
        "formula": "[C27H40O2]+",
        "rt": {"mean": 20.0},
    },
    {
        "formula": "[C24H26O12]+",
        "rt": {"mean": 30.0},
    },
]

simulated_sample_factory = SimulatedLCMSSampleFactory(adducts=adducts_spec) # type: ignore
sample = simulated_sample_factory(id="my_sample")
ms_data = MSData(sample)

mz = ms_data.get_spectrum(10).mz[0] # get the m/z value from the first feature


params = MakeChromatogramParameters(mz=[mz])
chrom = make_chromatograms(ms_data, params)[0]

fig, ax = plt.subplots()

ax.plot(chrom.time, chrom.int)
ax.set_xlabel("Rt [s]")
ax.set_ylabel("Intensity [au]")



"""Core metrics used by the data matrix."""

from typing import assert_never, cast

import numpy
from scipy.stats import median_abs_deviation, pearsonr, rankdata
from sklearn.decomposition import PCA

from ..enums import CorrelationMethod, NormalizationMethod, ScalingMethod
from .numpy import FloatArray, FloatArray1D, check_matrix_shape
from .transformation import normalize, scale


def cv(X: FloatArray, robust: bool = False) -> FloatArray1D:
    """Compute the :term:`CV` on the columns of a 2D array.

    The CV is computed as the quotient between the sample standard deviation and the sample mean.

    NaN values are ignored in the calculation.

    :param X: a 2D numpy array
    :param robust: If set to ``True`` uses the median and median absolute deviation instead of
        sample mean and sample standard deviation.
    :return: a 1D array with the CV of each column. If robust is set to ``False`` and less than two values in
        a column are not NaN, a NaN value will be returned. If all values in the column are zero or NaN, a NaN
        value will be obtained.

    """
    check_matrix_shape(X)
    x_std = _sample_std(X, robust=robust)
    x_mean = numpy.nanmedian(X, axis=0) if robust else numpy.nanmean(X, axis=0)
    return x_std / x_mean


def detection_rate(X: FloatArray, threshold: float | FloatArray = 0.0) -> FloatArray1D:
    """Compute the detection rate for all columns.

    The detection rate is defined as the fraction of elements in a column greater or equal than a threshold.
    NaN values are considered to be below the detection threshold.

    :param X: a 2D numpy array
    :param threshold: the detection threshold. If a scalar is provided, the same threshold is
        used for all columns. An array may be provided to use a different threshold for each
        column. If the threshold contains NaNs, the detection rate for the column will be set to zero.
    :return: a 1D with the detection rate of each column.

    """
    check_matrix_shape(X)

    n_rows, n_cols = X.shape

    if not isinstance(threshold, numpy.ndarray):
        threshold = numpy.ones(shape=(1, n_cols)) * threshold

    if threshold.size != n_cols:
        msg = "The threshold must be a scalar or a 1D float with size equal to the number of columns in X."
        raise ValueError(msg)

    threshold = threshold.reshape((1, threshold.size))

    non_detected_count = numpy.sum(numpy.isnan(X) | (X < threshold), axis=0)
    # set columns where threshold is nan to all undetected
    non_detected_count[numpy.isnan(threshold.flatten())] = n_rows

    return (n_rows - non_detected_count) / n_rows


def dratio(Xs: FloatArray, Xqc: FloatArray, robust: bool = False) -> FloatArray1D:
    r"""Compute the D-ratio metric for columns.

    The D-ratio is defined as the quotient between the standard deviation of QC data, or data that
    is expected to exhibit instrumental variation only and the standard deviation of sample data
    or data that presents `biological` variation.

    .. math::

        \textrm{D-Ratio} = \frac{S_{\textrm{QC}}}{S_{\textrm{sample}}}

    where :math:`S_{\textrm{sample}` is the sample standard deviation and :math:`S_{\textrm{QC}` is
    the QC standard deviation.

    NaN values in the sample or QC data will be ignored in the computation of the standard deviation.

    :param Xs: 2D array with sample data
    :param Xqc: 2D array with QC data
    :param robust: if set to ``True`` estimate the D-ratio using the median absolute deviation instead.
    :return: an 1D array with the D-ratio of each column. NaN values will be obtained if the sample standard
        deviation is zero. If the number of rows in `Xqc` or `Xs` is one, NaN values will also be obtained.

    """
    check_matrix_shape(Xs)
    check_matrix_shape(Xqc)
    return _sample_std(Xqc, robust=robust) / _sample_std(Xs, robust=robust)


def correlation(
    X: FloatArray,
    y: FloatArray1D,
    method: CorrelationMethod | str = CorrelationMethod.PEARSON,
) -> FloatArray1D:
    """Compute the correlation between the columns of a 2D array and a 1D array.

    :param X: the 2D array
    :param y: the 1D array
    :param method: ``"pearson"`` computes the Pearson's r coefficient. ``"spearman"``
        computes the spearman rank correlation coefficient.
    :return: the correlation coefficient between `y` and each `X` column.

    """
    if not isinstance(method, CorrelationMethod):
        method = CorrelationMethod(method)

    check_matrix_shape(X)

    if y.size != X.shape[0]:
        raise ValueError("The number of rows in `X` must match the length of `y`.")

    if y.size < 2:
        raise ValueError("At least a size of two is required to compute the correlation.")

    match method:
        case CorrelationMethod.PEARSON:
            return _corr_pearson(X, y)
        case CorrelationMethod.SPEARMAN:
            return _corr_spearman(X, y)
        case _ as never:
            assert_never(never)


def lod(X: FloatArray) -> FloatArray1D:
    """Estimates the limit of detection (LOD) of the data.

    :param X: a data matrix
    """
    check_matrix_shape(X)
    lod_ = numpy.nanmean(X, axis=0) + 3 * numpy.nanstd(X, axis=0, ddof=1)
    lod_[numpy.isnan(lod_)] = 0.0
    return lod_


def loq(X: FloatArray) -> FloatArray1D:
    """Estimates the limit of quantitation (LOQ) of the data.

    :param X: a data matrix
    """
    check_matrix_shape(X)
    lod_ = numpy.nanmean(X, axis=0) + 10 * numpy.nanstd(X, axis=0, ddof=1)
    lod_[numpy.isnan(lod_)] = 0.0
    return lod_


def pca(
    X: FloatArray,
    *,
    n_components=2,
    normalization=None,
    scaling=None,
    return_loadings=False,
    return_variance=False,
) -> tuple[FloatArray, FloatArray | None, FloatArray1D | None]:
    """Compute PCA on data.

    PCA requires that no NaN values

    :param X: the data array
    :param n_components: the number of Principal Components to compute.
    :param scaling: the scaling method applied to `X` columns. Refer to :py:class:`~tidyms2.core.enums.ScalingMethod`
        for the list of available scaling methods. If set to ``None``, no scaling is applied.
    :param normalization: One of the available normalization methods. Refer to
        :py:class:`~tidyms2.core.enums.NormalizationMethod` for the list of available normalization methods. If set
        to ``None``, do not perform row normalization.
    :param return_loadings: wether to return the PCA feature loadings.
    :pram return_variance: wether to return the PC variance.
    :return: A 2D array with PCA scores. If `return_loadings` is set to ``True`` also include a 2D array with
        the PCA loadings. If `return_pc_variance` is set to ``True`` also include a 1D array with PC variances.

    """
    check_matrix_shape(X)

    if isinstance(normalization, str):
        normalization = NormalizationMethod(normalization)

    if isinstance(scaling, str):
        scaling = ScalingMethod(scaling)

    if normalization is not None:
        X = normalize(X, normalization)

    if scaling is not None:
        X = scale(X, scaling)

    pca = PCA(n_components=n_components)
    scores = pca.fit_transform(X)
    n_components = pca.n_components_

    if return_loadings:
        loadings = pca.components_.T * numpy.sqrt(pca.explained_variance_)
    else:
        loadings = None

    if return_variance:
        variance = pca.explained_variance_
    else:
        variance = None

    return scores, loadings, variance


def _sample_std(X: FloatArray, robust: bool = False) -> FloatArray1D:
    """Compute an estimation of the sample standard deviation."""
    if robust:
        return median_abs_deviation(X, axis=0, scale="normal", nan_policy="omit")  # type: ignore
    return numpy.nanstd(X, axis=0, ddof=1)


def _corr_pearson(X: FloatArray, y: FloatArray1D) -> FloatArray1D:
    Y = numpy.tile(numpy.reshape(y, (y.size, 1)), X.shape[1])
    return cast(FloatArray1D, pearsonr(Y, X, axis=0)[0])


def _corr_spearman(X: FloatArray, y: FloatArray1D) -> FloatArray1D:
    X = rankdata(X, axis=0)
    y = rankdata(y)
    return _corr_pearson(X, y)



from itertools import product
from random import gauss

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from sklearn.cluster import DBSCAN

sample_size = [10, 20, 50, 100, 200, 500]
fractions = [0.1, 0.25, 0.5, 0.75, 1.0]
eps = [0.5, 1.0, 2.0, 3.0, 4.0]
n_reps = 5
results = list()

n_rows = 2
n_cols = 3
fig, ax = plt.subplots(
    nrows=n_rows,
    ncols=n_cols,
    sharex="all",
    sharey="all",
    figsize=(8, 6),
)

def add_jitter(x: float) -> float:
    """Add random perturbation to create a jitter effect in the plot."""
    return x + gauss(sigma=0.05)


def plot_for_sample_size(size: int, ax: Axes):
    """Create plot for a specific sample size"""
    sample_fraction_to_xy = dict()
    for _, f, e in product(range(n_reps), fractions, eps):
        noise_fraction = compute_noise_fraction(size, f, e)
        x, y = sample_fraction_to_xy.setdefault(f, (list(), list()))
        x.append(add_jitter(e))
        y.append(noise_fraction)
    for fraction, (x, y) in sample_fraction_to_xy.items():
        ax.scatter(x, y, edgecolors=None, s=10.0, label=fraction)


def compute_noise_fraction(size: int, fraction: float, eps: float):
    X = np.random.normal(size=(size, 2))
    min_samples = round(size * fraction)
    dbscan = DBSCAN(eps=eps, min_samples=min_samples, metric="chebyshev")
    dbscan.fit(X)
    cluster = dbscan.labels_
    return (cluster == -1).sum() / size

for k, size in enumerate(sample_size):
    i_row, i_col = divmod(k, 3)
    kax = ax[i_row, i_col]
    plot_for_sample_size(size, kax)

    kax.set_title(f"Sample size = {size}")
    kax.set_xticks([0, 1, 2, 3, 4, 5])

    if i_col == 0:
        kax.set_ylabel("Noise fraction")

    if i_row == 1:
        kax.set_xlabel("eps")

    if i_col == 2 and i_row == 1:
        kax.legend(title="sample fraction")


from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from tidyms2.core.models import Sample
from tidyms2.simulation.lcms import SimulatedLCMSAdductSpec, simulate_data_matrix

formula_list = ["[C54H104O6]+", "[C27H40O2]+", "[C24H26O12]+"]
rt_list = [10.0, 20.0, 30.0]
abundances = [1000.0, 2000.0, 3000.0]
adducts = list()
for rt, abundance, formula in zip(rt_list, abundances, formula_list):
    adduct = SimulatedLCMSAdductSpec(
        formula=formula,
        abundance={
            "group-a": {"mean": abundance, "std": 50.0},
            "group-b": {"mean": abundance * 2, "std": 50.0}
        },  # type: ignore
    )
    adducts.append(adduct)


samples = list()
for k in range(10):
    id_ = f"sample-{k}"
    path = Path.cwd() / f"{id_}.mzML"
    s = Sample(id=id_, path=path)
    s.meta.order = k
    s.meta.group = "group-a" if k % 2 else "group-b"
    samples.append(s)

data = simulate_data_matrix(adducts, samples)

col = data.get_columns(0)[0]
order = np.array([s.meta.order for s in data.list_samples()])
group = np.array([s.meta.group for s in data.list_samples()])

fig, ax = plt.subplots()

mask_group_a = group == "group-a"
mask_group_b = group == "group-b"
ax.scatter(order[mask_group_a], col.data[mask_group_a], label="group-a")
ax.scatter(order[mask_group_b], col.data[mask_group_b], label="group-b")
ax.legend()
ax.set_title(f"mz={col.feature.mz:.4f}, Rt={col.feature.descriptors["rt"]:.1f} s")
ax.set_xlabel("Run order")
ax.set_ylabel("Intensity [au]")
ax.set_ylim(0, 1200)



import pytest

from tidyms2.storage.memory import OnMemorySampleStorage

from ... import helpers
from ...helpers import ConcreteFeature, ConcreteRoi


@pytest.fixture
def sample_storage(tmp_path) -> OnMemorySampleStorage[ConcreteRoi, ConcreteFeature]:
    sample = helpers.create_sample(tmp_path, 1)
    return OnMemorySampleStorage(sample, ConcreteRoi, ConcreteFeature)


@pytest.fixture
def sample_storage_with_rois(sample_storage):
    roi_extractor = helpers.DummyRoiExtractor()
    roi_extractor.apply(sample_storage)
    return sample_storage


@pytest.fixture
def sample_storage_with_features(sample_storage_with_rois):
    feature_extractor = helpers.DummyFeatureExtractor()
    feature_extractor.apply(sample_storage_with_rois)
    return sample_storage_with_rois



"""Functions to calculate compatible molecular formulas within a given tolerance."""

from __future__ import annotations

import numpy as np
import pydantic

from ..core.utils.numpy import IntArray, cartesian_product_from_ranges
from .atoms import Isotope
from .config import FormulaGeneratorConfiguration
from .context import PTABLE

# name conventions used:
# M is used for molecular mass
# m is used for nominal mass
# d is used for mass defect
# e.g. for (12C)(1H)2(16O)2, M = 46.0055; m = 46; d = 0.0055
# dp is used for the contribution to d of isotopes with positive mass defect
# dn is used for the contribution to d of isotopes with negative mass defect
# d = dp + dn
# for (12C)(1H)2(16O)2 dp = 0.0157, dn = -0.0102.
# q and r are used for the quotient and remainder of the division of m by 12
# for (12C)(1H)2(16O)2 q = 3; r = 10
# mp is used for the contribution to m of isotopes with positive mass defect
# mn is used for the contribution to m of isotopes with negative mass defect
# qp and rp are the quotient and remainder of the division of mp by 12
# qn and rn are the quotient and remainder of the division of mn by 12
# mc is used fot the contribution to m of 12C.
# qc is the quotient of division of mc by 12.
# rc is not used but it is always zero.


class FormulaGenerator:
    """Generate sum formulas based on exact mass values.

    Refer to the :ref:`user guide <generating-formulas-guide>` for usage instructions.

    """

    def __init__(self, config: FormulaGeneratorConfiguration):
        """FormulaGenerator constructor.

        :param config: The formula generator configuration.

        """
        bounds = FormulaCoefficientBounds.from_isotope_str(config.bounds)
        self.bounds = bounds.bounds_from_mass(config.max_M, 0.0)

        dp_min, dp_max, dn_min, dn_max = self.bounds.get_defect_bounds()

        self._min_defect = dn_min + dp_min
        self._max_defect = dp_max + dn_max

        c12 = PTABLE.get_isotope("12C")
        self._has_carbon = c12 in self.bounds.bounds
        self._query = None
        self._results = None
        self._n_results = None

        # build Coefficients
        pos_restrictions, neg_restrictions = self.bounds.split_pos_neg()
        # add a dummy negative (or positive) isotope when only positive
        # (negative) isotopes are being used. This is a hack that prevents
        # making changes in the formula generation code...
        _add_dummy_isotope(pos_restrictions, neg_restrictions)
        self.pos = pos_restrictions.make_coefficients(config.max_M, reverse=True)
        self.neg = neg_restrictions.make_coefficients(config.max_M)

    def __repr__(self):
        return f"FormulaGenerator(bounds={self.bounds.bounds})"

    def get_results(self) -> dict:
        """Retrieve the formula generator results in the internal format.

        :return: a mapping of nominal masses of the results to a tuple of three arrays:
            1. the row index of positive coefficients.
            2. the row index of negative coefficients.
            3. the number of 12C in the formula.

        """
        if self._results is None:
            raise ValueError("No results found. This method must be called after `generate_formulas`")
        return self._results

    def get_n_results(self) -> int:
        """Retrieve the number of formulas found after a query."""
        if self._n_results is None:
            raise ValueError("No results found. This method must be called after `generate_formulas`")
        return self._n_results

    def generate_formulas(
        self, M: float, tolerance: float, min_defect: float | None = None, max_defect: float | None = None
    ):
        """Compute formulas compatibles with the given query mass.

        The formulas are computed assuming neutral species. If charged species are used, mass
        values must be corrected using the electron mass.

        Results are stored in an internal format, use :py:func:`results_to_array` to
        obtain the compatible formulas.

        :param M: query mass used for formula generation
        :param tolerance: mass tolerance to search compatible formulas
        :param min_defect: if provided, filter formulas with mass defects lower than this value
        :param max_defect: if provided, filter formulas with mass defects greater than this value

        >>> from tidyms2.chem import FormulaGenerator
        >>> fg_bounds = {"C": (0, 5), "H": (0, 10), "O": (0, 4)}
        >>> fg = FormulaGenerator(fg_bounds)
        >>> fg.generate_formulas(46.042, 0.005)

        """
        if M <= 0.0:
            msg = "`M` must be a positive number. Got {}".format(M)
            raise ValueError(msg)

        if tolerance <= 0.0:
            msg = "`tolerance` must be a positive number. Got {}".format(tolerance)
            raise ValueError(msg)

        min_defect = self._min_defect if min_defect is None else min_defect
        max_defect = self._max_defect if max_defect is None else max_defect

        self._results, self._n_results = _generate_formulas(
            M, tolerance, self.bounds, self.pos, self.neg, min_defect=min_defect, max_defect=max_defect
        )

    def results_to_array(self) -> tuple[np.ndarray, list[Isotope], np.ndarray]:
        """Convert results to an array of coefficients.

        :return: tuple containing a 2D array wit rows of formula coefficients, a list
            of isotopes associated with each coefficient and a 1D array with the exact
            mass of each formula.

        >>> import tidyms as ms
        >>> fg_bounds = {"C": (0, 5), "H": (0, 10), "O": (0, 4)}
        >>> fg = ms.chem.FormulaGenerator(fg_bounds)
        >>> fg.generate_formulas(46.042, 0.005)
        >>> coeff, isotopes, M = fg.results_to_array()

        """
        if self._results is None or self._n_results is None:
            raise ValueError("No results computed.")
        return _results_to_array(self._results, self._n_results, self.pos, self.neg, self._has_carbon)


class IsotopeCoeffBounds(pydantic.BaseModel):
    """Store a lower and upper bounds for formula coefficients bounds."""

    lower: pydantic.NonNegativeInt
    upper: pydantic.NonNegativeInt


class MassBounds(pydantic.BaseModel):
    """Store lower and upper bounds for mass."""

    lower: float
    upper: float


class MassDefectBounds(pydantic.BaseModel):
    """Store positive and negative mass defect bounds build from formula coefficients."""

    pos: MassBounds
    neg: MassBounds


class FormulaCoefficientBounds(pydantic.BaseModel):
    """Mapping from isotopes to upper and lower bounds."""

    bounds: dict[Isotope, IsotopeCoeffBounds]
    """A mapping from isotopes to coefficient bounds"""

    def __getitem__(self, item):
        return self.bounds[item]

    def bounds_from_mass(self, M: float, tol: float) -> FormulaCoefficientBounds:
        """Compute the mass-based bounds for each isotope.

        The bounds are refined using the values for each isotope.

        """
        bounds = dict()
        for i, b in self.bounds.items():
            lower = max(0, b.lower)
            upper = min(int((M + tol) / i.m), b.upper)
            bounds[i] = IsotopeCoeffBounds(lower=lower, upper=upper)
        return FormulaCoefficientBounds(bounds=bounds)

    def bound_negative_positive_defect(self, defect: float, tolerance: float) -> MassDefectBounds:
        """Bound positive and negative contributions to mass defect.

        :param defect: mass defect value of the Query.
        :param tolerance: mass tolerance

        """
        min_p, max_p, min_n, max_n = self.get_defect_bounds()
        max_pos = float(min(defect - min_n, max_p) + tolerance)
        min_neg = float(max(defect - max_p, min_n) - tolerance)
        min_pos = float(max(defect - max_n, min_p) - tolerance)
        max_neg = float(min(defect - min_p, max_n) + tolerance)
        return MassDefectBounds(
            pos=MassBounds(lower=min_pos, upper=max_pos), neg=MassBounds(lower=min_neg, upper=max_neg)
        )

    def get_nominal_defect_candidates(self, M: float, tol: float) -> tuple[list[int], list[float]]:
        """Split mass into possible values of nominal mass and mass defect."""
        dp_min, dp_max, dn_min, dn_max = self.get_defect_bounds()
        m_min = int(M - tol - dn_max - dp_max) + 1
        m_max = int(M + tol - dp_min - dn_min) + 1
        m_candidates = list(range(m_min, m_max))
        d_candidates = [M - x for x in m_candidates]
        return m_candidates, d_candidates

    def split_pos_neg(self) -> tuple[FormulaCoefficientBounds, FormulaCoefficientBounds]:
        """Split bounds into a two new instances, one with positive defect isotopes and other with negative defects."""
        pos_bounds = dict()
        neg_bounds = dict()
        for isotope, bounds in self.bounds.items():
            defect = isotope.d
            if defect > 0:
                pos_bounds[isotope] = bounds
            elif defect < 0:
                neg_bounds[isotope] = bounds
        pos = FormulaCoefficientBounds(bounds=pos_bounds)
        neg = FormulaCoefficientBounds(bounds=neg_bounds)
        return pos, neg

    def get_defect_bounds(self) -> tuple[float, float, float, float]:
        """Compute the minimum and maximum value of the mass defect."""
        min_positive, max_positive, min_negative, max_negative = 0, 0, 0, 0
        for isotope, isotope_bounds in self.bounds.items():
            defect = isotope.d
            min_tmp = defect * isotope_bounds.lower
            max_tmp = defect * isotope_bounds.upper
            if defect > 0:
                min_positive += min_tmp
                max_positive += max_tmp
            else:
                min_negative += max_tmp
                max_negative += min_tmp
        return min_positive, max_positive, min_negative, max_negative

    def make_coefficients(self, max_M: float, reverse: bool = False) -> FormulaCoefficients:
        """Generate coefficients for FormulaGenerator."""
        return FormulaCoefficients(self, max_M, True, reverse)

    @staticmethod
    def from_isotope_str(bounds: dict[str, tuple[int, int]]) -> FormulaCoefficientBounds:
        """Create a _Bounds instance from a list of isotope strings."""
        res = dict()
        for i, (lb, ub) in bounds.items():
            try:
                element = PTABLE.get_element(i)
                isotope = element.mmi
            except KeyError:
                isotope = PTABLE.get_isotope(i)
            res[isotope] = IsotopeCoeffBounds(lower=lb, upper=ub)
        return FormulaCoefficientBounds(bounds=res)


class FormulaCoefficients:
    """Named tuple with elements with positive/negative mass defect.

    :param coefficients: np.array[int]
        Formula coefficients. Each row is a formula, each column is an isotope.
    :param isotopes : List[Isotopes]
        element associated to each column of coefficients.
    :param M: monoisotopic mass associated to each row of coefficients.
    :param q: quotient between the nominal mass and 12.
    :param r_to_index: Maps remainders of division of m by 12 to rows of `coefficients`.
    :param r_to_d: Maps remainders of division of m by 12 to mass defect  values. Each
        value match to the corresponding index in `r_to_index`.

    """

    def __init__(
        self,
        bounds: FormulaCoefficientBounds,
        max_mass: float,
        return_sorted: bool,
        return_reversed: bool,
    ):
        self.isotopes = list(bounds.bounds)
        i_M = np.array([isotope.m for isotope in self.isotopes])
        i_m = np.array([isotope.a for isotope in self.isotopes])
        i_d = np.array([isotope.d for isotope in self.isotopes])

        # create coefficients array
        range_list = [list(range(b.lower, b.upper + 1)) for b in bounds.bounds.values()]
        coefficients = cartesian_product_from_ranges(*range_list)

        # sort coefficients by mass defect
        d = np.matmul(coefficients, i_d)
        if return_sorted:
            sorted_index = np.argsort(d)
            if return_reversed:
                sorted_index = sorted_index[::-1]
            coefficients = coefficients[sorted_index, :]
            d = d[sorted_index]

        # remove coefficients with mass higher than the maximum mass
        M = np.matmul(coefficients, i_M)
        valid_M = M <= max_mass
        coefficients = coefficients[valid_M, :]
        d = d[valid_M]
        M = M[valid_M]

        # Compute nominal mass, quotient and remainder
        m = np.matmul(coefficients, i_m)
        q, r = np.divmod(m, 12)

        # group mass defects and coefficient row index by remainder value
        r_to_d = _make_remainder_arrays(d, r)
        r_to_index = _make_remainder_arrays(np.arange(d.size), r)

        self.M = M
        self.coefficients = coefficients
        self.q = q
        self.r_to_index = r_to_index
        self.r_to_d = r_to_d


class MassQuery(pydantic.BaseModel):
    """Stores values for a Mass query."""

    m: pydantic.PositiveInt
    """nominal mass of the query."""

    d: float
    """mass defect of the query."""

    q: pydantic.NonNegativeInt
    """quotient of the division of `m` by 12."""

    r: pydantic.NonNegativeInt
    """remainder of the division of `m` by 12."""

    tol: pydantic.PositiveFloat
    """Mass tolerance for the query"""

    carbon_bounds: IsotopeCoeffBounds
    """Minimum and maximum number of 12C in the formula"""

    defect_bounds: MassDefectBounds
    """Mass defect bounds for valid formulas"""

    @classmethod
    def from_coeffs(cls, m: int, d: float, tol: float, bounds: FormulaCoefficientBounds):
        """Create a mass query from formula bounds."""
        q, r = divmod(m, 12)
        d_bounds = bounds.bound_negative_positive_defect(d, tol)

        c12 = PTABLE.get_isotope("12C")
        if c12 in bounds.bounds:
            c12_bounds = bounds.bounds[c12].model_copy()
        else:
            c12_bounds = IsotopeCoeffBounds(lower=0, upper=0)

        return MassQuery(m=m, d=d, q=q, r=r, defect_bounds=d_bounds, carbon_bounds=c12_bounds, tol=tol)


def _add_dummy_isotope(pos: FormulaCoefficientBounds, neg: FormulaCoefficientBounds):
    """Add dummy isotopes to empty bounds.

    This is to positive or negative elements to solve the mass defect problem in cases that there aren't any
    positive/negative isotopes.

    """
    if len(pos.bounds) == 0:
        h1 = PTABLE.get_isotope("1H")
        pos.bounds[h1] = IsotopeCoeffBounds(lower=0, upper=0)

    if len(neg.bounds) == 0:
        o16 = PTABLE.get_isotope("16O")
        neg.bounds[o16] = IsotopeCoeffBounds(lower=0, upper=0)


def _results_to_array(
    results: dict,
    n: int,
    pos: FormulaCoefficients,
    neg: FormulaCoefficients,
    has_carbon: bool,
) -> tuple[np.ndarray, list[Isotope], np.ndarray]:
    """Convert results from guess_formula to a numpy array of coefficients."""
    isotopes = pos.isotopes + neg.isotopes
    if has_carbon:
        isotopes = [PTABLE.get_isotope("12C")] + isotopes
        ic = 0
        i_pos = 1
        i_neg = i_pos + len(pos.isotopes)
    else:
        ic = 0
        i_pos = 0
        i_neg = i_pos + len(pos.isotopes)
    res = np.zeros((n, len(isotopes)), dtype=int)
    mass = np.zeros(n)
    start = 0
    for k in results:
        end = start + results[k][0].size
        res[start:end, i_pos:i_neg] = pos.coefficients[results[k][0], :]
        res[start:end, i_neg:] = neg.coefficients[results[k][1], :]
        if has_carbon:
            res[start:end, ic] = results[k][2]
            mass[start:end] = pos.M[results[k][0]] + 12 * np.array(results[k][2]) + neg.M[results[k][1]]
        else:
            mass[start:end] = pos.M[results[k][0]] + neg.M[results[k][1]]
        start = end
    res, isotopes = _remove_dummy_isotopes(pos, neg, has_carbon, isotopes, res)
    return res, isotopes, mass


def _remove_dummy_isotopes(
    pos: FormulaCoefficients,
    neg: FormulaCoefficients,
    has_carbon: bool,
    isotopes: list[Isotope],
    res: IntArray,
) -> tuple[IntArray, list[Isotope]]:
    has_pos = pos.coefficients.max() > 0
    has_neg = neg.coefficients.max() > 0

    if not has_pos:
        pop_ind = 1 if has_carbon else 0
        isotopes.pop(pop_ind)
        res = np.delete(res, pop_ind, axis=1)

    if not has_neg:
        pop_ind = res.shape[1] - 1
        isotopes.pop(pop_ind)
        res = np.delete(res, pop_ind, axis=1)

    return res, isotopes


def _make_remainder_arrays(x, r) -> dict[int, np.ndarray]:
    """Create a dictionary where each key is a value of r and the values are the corresponding x values.

    Auxiliary function of _make_coefficients.

    :param x: array
    :param r: array of remainders.

    Returns
    -------
    r_to_x : Dict

    """
    r_to_x = dict()
    for k in range(12):
        xk = x[r == k]
        r_to_x[k] = xk
    return r_to_x


def _generate_formulas(
    M: float,
    tol: float,
    bounds: FormulaCoefficientBounds,
    pos: FormulaCoefficients,
    neg: FormulaCoefficients,
    min_defect: float,
    max_defect: float,
):
    """Find formulas compatible with a given mass."""
    bounds = bounds.bounds_from_mass(M, tol)
    # possible values of nominal mass and mass defect based on the coefficient bounds.
    m_candidates, d_candidates = bounds.get_nominal_defect_candidates(M, tol)
    res = dict()
    n = 0  # number of valid formulas
    for m, d in zip(m_candidates, d_candidates):
        if (d + tol < min_defect) or (d - tol > max_defect):
            continue

        query = MassQuery.from_coeffs(m, d, tol, bounds)

        pos_index = list()
        neg_index = list()
        qc = list()
        for i in range(12):
            results = _generate_formulas_i(i, query, pos, neg)
            if results is not None:
                pos_index_i, neg_index_i, qc_i = results
                n += pos_index_i.size
                pos_index.append(pos_index_i)
                neg_index.append(neg_index_i)
                qc.append(qc_i)
        if pos_index:
            res[m] = (np.hstack(pos_index), np.hstack(neg_index), np.hstack(qc))
    return res, n


def _generate_formulas_i(
    i: int,
    query: MassQuery,
    pos: FormulaCoefficients,
    neg: FormulaCoefficients,
) -> tuple[np.ndarray, np.ndarray, np.ndarray] | None:
    """Solve the mass defect problem for fixed rp and rn values.

    Auxiliary function to _generate_formulas.
    """
    # solves the mass defect problem for fixe rp and rn values.
    # Finds all positive coefficients with rp == i and their matching negative
    # such that |d - dp - dn | <= tol.
    # These values are filtered taking into account the number of 12C, qc, as
    # follows: min_nc <= qc <= max_nc
    # the results are organized into three arrays
    # p_index contains the index to a row of pos.coeff
    # n_index contains the index to a row of neg.coeff
    # q_c contains the number of 12C in the formula

    rp = i
    rn = (query.r - i) % 12

    # find valid positive mass defect values
    rel_dp = query.d - pos.r_to_d[rp]
    rel_dp_bounds = query.d - query.defect_bounds.pos.upper, query.d - query.defect_bounds.pos.lower
    p_start, p_end = np.searchsorted(rel_dp, rel_dp_bounds)
    rel_dp = rel_dp[p_start:p_end]

    # filter values based on valid number of C
    p_index = pos.r_to_index[rp][p_start:p_end]
    qp = pos.q[p_index]
    valid_qp = (query.q - qp) >= query.carbon_bounds.lower
    rel_dp = rel_dp[valid_qp]
    p_index = p_index[valid_qp]

    # find valid negative mass defect values
    dn = neg.r_to_d[rn]
    n_index = neg.r_to_index[rn]
    n_start = np.searchsorted(dn, rel_dp - query.tol)
    n_end = np.searchsorted(dn, rel_dp + query.tol)

    # create three arrays, where each element corresponds to an index of valid
    # positive coeff, negative coeff and number of 12C atoms.
    n_index_size = n_end - n_start
    valid_dn = n_index_size > 0
    n_start = n_start[valid_dn]
    n_end = n_end[valid_dn]
    n_index_size = n_index_size[valid_dn]
    p_index = p_index[valid_dn]
    if p_index.size:
        p_index = np.repeat(p_index, n_index_size)
        n_index = np.hstack([n_index[s:e] for s, e in zip(n_start, n_end)])

        qp = pos.q[p_index]
        qn = neg.q[n_index]
        extra_c = int((rp + rn) >= 12)
        qc = query.q - qp - qn - extra_c

        # valid results
        valid_qc = (qc >= query.carbon_bounds.lower) & (qc <= query.carbon_bounds.upper)
        p_index = p_index[valid_qc]
        n_index = n_index[valid_qc]
        qc = qc[valid_qc]
        if p_index.size:
            results = p_index, n_index, qc
        else:
            results = None
    else:
        results = None
    return results



"""mzML reader implementation.

Rationale
---------

mzML files can have typical sizes that span from 50 MiB to more than 1000 MiB.
mzML files are basically xml files with the acquired data and metadata
associated with the acquisition conditions. The data is inside <spectrum> and
<chromatogram> elements.  From the expected file sizes, it is clear that
loading the entire file in memory is not desirable.
There are tools to parse xml files in an incremental way (such as iterparse
inside the xml module), but the problem with this approach is that accessing a
specific scan in the file turns out to be rather slow.
Indexed mzML files have an <indexedmzML> tag that allows to search in a fast
way the location of spectrum and chromatogram tags.
Taking advantage of this information, this module search spectrum and
chromatogram elements and only loads the part associated with the selected
spectrum or chromatogram.
The function build_offset_list creates the list of offsets associated with
chromatograms and spectra using the information in <indexedmzML>. In the
case of non-indexed files, the offset list is built from scratch.
The functions get_spectrum and get_chromatogram takes the offset information
and extracts the relevant data from each spectrum/chromatogram.

# mzML specification:
# https://www.psidev.info/mzML
# terms defined here:
# https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo

build_offset_list : Finds offsets with the location of spectra and chromatograms
get_spectrum : Extracts data from a spectrum using offsets.
get_chromatogram : Extracts data from a chromatogram using offsets.

"""

import base64
import enum
import re
import zlib
from os import SEEK_END
from os.path import getsize
from pathlib import Path
from typing import Any, cast
from xml.etree.ElementTree import Element, fromstring

import numpy as np

from ..core.models import Chromatogram, MSSpectrum, Sample
from .msdata import reader_registry


@reader_registry.register("mzML", "mzml")
class MZMLReader:
    """mzML format reader."""

    def __init__(self, src: Path | Sample, rebuild_index: bool = False):
        self.path = src if isinstance(src, Path) else src.path
        sp_offset, chrom_offset, index_offset = _build_offset_list(self.path, rebuild_index)
        self.spectra_offset = sp_offset
        self.chromatogram_offset = chrom_offset
        self.index_offset = index_offset

        self.n_chromatograms = len(self.chromatogram_offset)
        self.n_spectra = len(self.spectra_offset)

    def get_spectrum(self, index: int) -> MSSpectrum:
        """Retrieve a spectrum from file."""
        return _get_spectrum(
            self.path,
            self.spectra_offset,
            self.chromatogram_offset,
            self.index_offset,
            index,
        )

    def get_n_spectra(self) -> int:
        """Return the number of spectra in the file."""
        return self.n_spectra

    def get_n_chromatograms(self) -> int:
        """Return the number of chromatogram in the file."""
        return self.n_chromatograms

    def get_chromatogram(self, index: int) -> Chromatogram:
        """Retrieve a chromatogram from file."""
        return _get_chromatogram(
            self.path,
            self.spectra_offset,
            self.chromatogram_offset,
            self.index_offset,
            index,
        )


UNSUPPORTED_COMPRESSION = frozenset(
    (
        "MS:1002312",
        "MS:1002313",
        "MS:1002314",
        "MS:1002746",
        "MS:1002747",
        "MS:1002748",
        "MS:1003088",
        "MS:1003089",
        "MS:1003090",
    )
)


class _CompressionType(str, enum.Enum):
    """The compression types defined in the mzML spec and their accession numbers."""

    ZLIB = "MS:1000574"


MS_LEVEL = "MS:1000511"


class _ArrayType(str, enum.Enum):
    """The array types defined in the mzML specification and their accession numbers."""

    MZ_ARRAY = "MS:1000514"
    """An array with m/z data"""

    INT_ARRAY = "MS:1000515"
    """An array with spectral intensity"""

    TIME_ARRAY = "MS:1000595"
    """An array with with time data"""

    TIME = "MS:1000016"  # time value in scans


class _Polarity(str, enum.Enum):
    """mzML polarity values."""

    NEGATIVE = "MS:1000129"
    POSITIVE = "MS:1000130"


class _TimeUnits(str, enum.Enum):
    """mzML time units."""

    SECONDS = "UO:0000010"
    MINUTES = "UO:0000031"


class _DataTypes(str, enum.Enum):
    FLOAT16 = "UNKNOWN"
    FLOAT32 = "MS:1000521"
    FLOAT64 = "MS:1000523"
    INT32 = "MS:1000519"
    INT64 = "MS:1000522"


_MZML_TYPE_TO_NUMPY_TYPE = {
    _DataTypes.FLOAT16.value: np.float16,
    _DataTypes.FLOAT32.value: np.float32,
    _DataTypes.FLOAT64.value: np.float64,
    _DataTypes.INT32.value: np.int32,
    _DataTypes.INT64.value: np.int64,
}


class MZMLError(ValueError):
    """Exception raised when an error happens while parsing an mzML file."""


def _build_offset_list(filename: Path, rebuild_index: bool) -> tuple[list[int], list[int], int]:
    """Find the offset values where Spectrum or Chromatogram elements start.

    :param filename: path to a mzML file
    :param rebuild_index: if set to ``True`` rebuilds the index even if the file is indexed

    :return: Offsets where spectrum element start, offsets where chromatogram elements start
        Offset where the index starts. If the file is not indexed, return the file size.

    """
    if not rebuild_index and is_indexed(filename):
        index_offset = _get_index_offset(filename)
        spectra_offset, chromatogram_offset = _build_offset_list_indexed(filename, index_offset)
    else:
        spectra_offset, chromatogram_offset = _build_offset_list_non_indexed(filename)
        index_offset = getsize(filename)
    return spectra_offset, chromatogram_offset, index_offset


def _get_spectrum(
    filename: Path, spectra_offset: list[int], chromatogram_offset: list[int], index_offset: int, n: int
) -> MSSpectrum:
    """Extract spectrum data from file.

    :param filename: path to mzML file.
    :param :spectra_offset : offset list obtained from `_build_offset_list`.
    :parm chromatogram_offset: Offset list obtained from `_build_offset_list`.
    :parma index_offset: offset obtained from `_build_offset_list`.
    :param n: scan number to select.
    :returns: dictionary with m/z, intensity, polarity , time, and ms level.

    """
    xml_str = _get_xml_data(filename, spectra_offset, chromatogram_offset, index_offset, n, "spectrum")
    elements = list(fromstring(xml_str))
    spectrum = dict()
    spectrum["index"] = n
    for el in elements:
        tag = el.tag
        if tag == "cvParam":
            accession = el.attrib.get("accession")
            if accession == MS_LEVEL:
                ms_level = el.attrib.get("value")
                spectrum["ms_level"] = 1 if ms_level is None else int(ms_level)
            elif accession == _Polarity.NEGATIVE:
                spectrum["polarity"] = -1
            elif accession == _Polarity.POSITIVE:
                spectrum["polarity"] = 1
        elif tag == "scanList":
            spectrum["time"] = _get_time(el)
        elif tag == "binaryDataArrayList":
            sp_data = _parse_binary_data_list(el)
            spectrum.update(sp_data)

    return MSSpectrum(**spectrum)


def _get_chromatogram(
    filename: Path, spectra_offset: list[int], chromatogram_offset: list[int], index_offset: int, n: int
) -> Chromatogram:
    """Extract time and intensity from xml chunk.

    :parm filename: path to mzML file
    :param spectra_offset: offset list obtained from _build_offset_list
    :param chromatogram_offset: offset list obtained from _build_offset_list
    :param index_offset: offset obtained from _build_offset_list
    :param n: chromatogram number
    :returns: dict

    """
    xml_str = _get_xml_data(filename, spectra_offset, chromatogram_offset, index_offset, n, "chromatogram")
    elements = fromstring(xml_str)
    name = elements.attrib.get("id")
    chromatogram: dict[str, Any] = dict(name=name)
    for el in list(elements):
        tag = el.tag
        if tag == "binaryDataArrayList":
            chrom_data = _parse_binary_data_list(el)
            chromatogram.update(chrom_data)
    return Chromatogram(**chromatogram)


class _ReverseReader:
    """Read file objects starting from the EOF.

    :param filename: Path to the file
    :param buffer_size: size of the chunk to get when using the `read_chunk` method
    :param kwargs: keyword arguments to pass to the open function

    """

    def __init__(self, filename: Path, buffer_size: int, **kwargs):
        self.file = open(filename, **kwargs)
        self.buffer_size = buffer_size
        self._size = self.file.seek(0, SEEK_END)
        self.offset = self._size
        self._is_complete = False

    @property
    def offset(self) -> int:
        """Retrieve the file offset."""
        return self._offset

    @offset.setter
    def offset(self, value: int):
        if value < 0:
            value = 0
            self._is_complete = True
        self._offset = value
        self.file.seek(value)

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.file.close()

    def read_chunk(self) -> str | None:
        """Read a chunk of data from the file, starting from the end.

        If the beginning of the file has been reached, it returns None.
        """
        if self._is_complete:
            res = None
        else:
            self.offset = self.offset - self.buffer_size
            res = self.file.read(self.buffer_size)
        return res

    def reset(self):
        """Set the position of the reader to the EOF."""
        self.offset = self._size
        self._is_complete = False


def _read_binary_data_array(element: Element) -> tuple[np.ndarray, str]:
    """Extract the binary data and data kind from a binaryArray element.

    data : array
    # kind : can be one of {"mz", "spint", "time"}

    """
    has_zlib_compression = False
    data = None
    kind = "none"
    units = None
    dtype = None
    for e in element:
        tag = e.tag
        if tag == "binary":
            data = e.text if e.text is None else e.text.strip()
        elif tag == "cvParam":
            accession = e.attrib.get("accession")
            if accession in UNSUPPORTED_COMPRESSION:
                msg = "Currently only zlib compression is supported."
                raise MZMLError(msg)
            elif accession == _CompressionType.ZLIB:
                has_zlib_compression = True
            elif accession == _ArrayType.INT_ARRAY:
                kind = "int"
            elif accession == _ArrayType.MZ_ARRAY:
                kind = "mz"
            elif accession == _ArrayType.TIME_ARRAY:
                kind = "time"
                units = e.attrib.get("unitAccession")
            if accession in _MZML_TYPE_TO_NUMPY_TYPE:
                dtype = _MZML_TYPE_TO_NUMPY_TYPE[accession]
    if data:
        data = base64.b64decode(data)
        if has_zlib_compression:
            data = zlib.decompress(data)
        data = np.frombuffer(data, dtype=dtype).copy()
        if kind == "time":
            data = _time_to_seconds(data, units)
    else:
        data = np.array([])
    return data, kind


def _parse_binary_data_list(element: Element) -> dict:
    """Extract the data from a binaryDataArrayList.

    :param element: Element
    :returns: dictionary that maps data kind to its correspondent data array.

    """
    res = dict()
    for e in element:
        data, data_type = _read_binary_data_array(e)
        res[data_type] = data
    return res


def is_indexed(filename: Path) -> bool:
    """Check if a mzML file is indexed.

    :param filename: Path

    Notes
    -----
    This function assumes that the mzML file is validated. It looks up the last
    closing tag, that should be </indexedmzML> if the file is indexed.

    """
    with _ReverseReader(filename, 1024, mode="r") as fin:
        end_tag = "</indexedmzML>"
        chunk = cast(str, fin.read_chunk())
        res = chunk.find(end_tag) != -1
    return res


def _get_index_offset(filename: Path) -> int:
    """Search the byte offset where the indexListOffset element starts.

    :param filename: Path

    :returns: index offset

    """
    tag = "<indexListOffset>"
    # reads mzml backwards until the tag is found
    # we exploit the fact that according to the mzML schema the
    # indexListOffset should have neither attributes nor sub elements
    with _ReverseReader(filename, 1024, mode="r") as fin:
        xml = ""
        ind = -1
        while ind == -1:
            chunk = fin.read_chunk()
            if chunk is not None:
                xml = chunk + xml
                ind = chunk.find(tag)
        # starts at the beginning of the text tag
        start = ind + len(tag)
        xml = xml[start:]
        end = xml.find("<")
    index_offset = int(xml[:end])
    return index_offset


def _build_offset_list_non_indexed(filename: Path) -> tuple[list[int], list[int]]:
    """Build manually the indices for non-indexed mzML files."""
    # indices are build by finding the offset where spectrum or chromatogram
    # elements starts.
    spectrum_regex = re.compile("<spectrum .[^(><.)]+>")
    chromatogram_regex = re.compile("<chromatogram .[^(><.)]+>")
    ind = 0
    spectrum_offset_list = list()
    chromatogram_offset_list = list()
    with open(filename) as fin:
        while True:
            line = fin.readline()
            if line == "":
                break
            spectrum_offset = _find_spectrum_tag_offset(line, spectrum_regex)
            if spectrum_offset:
                spectrum_offset_list.append(ind + spectrum_offset)
            chromatogram_offset = _find_chromatogram_tag_offset(line, chromatogram_regex)
            if chromatogram_offset:
                chromatogram_offset_list.append(ind + chromatogram_offset)
            ind += len(line)
    return spectrum_offset_list, chromatogram_offset_list


def _find_spectrum_tag_offset(line: str, regex: re.Pattern) -> int | None:
    if line.lstrip().startswith("<spectrum"):
        match = regex.search(line)
        if match:
            start, end = match.span()
        else:
            start = None
        return start


def _find_chromatogram_tag_offset(line: str, regex: re.Pattern) -> int | None:
    if line.lstrip().startswith("<chromatogram"):
        match = regex.search(line)
        if match:
            start, end = match.span()
        else:
            start = None
        return start


def _build_offset_list_indexed(filename: Path, index_offset: int) -> tuple[list[int], list[int]]:
    """Build a list of offsets where spectra and chromatograms are stored.

    :param filename: Path
    :param index_offset : offset obtained from _get_index_offset
    :returns: offset where spectra are stored and offset where chromatograms are stored

    """
    end_tag = "</indexList>"
    with open(filename, "r") as fin:
        fin.seek(index_offset)
        index_xml = fin.read()
        end = index_xml.find(end_tag)
        index_xml = index_xml[: end + len(end_tag)]
    index_xml = fromstring(index_xml)

    spectra_offset = list()
    chromatogram_offset = list()
    for index in index_xml:
        for offset in index:
            if offset.text is None:
                raise MZMLError("Non integer offset found in indexed file.")
            value = int(offset.text)
            if index.attrib["name"] == "spectrum":
                spectra_offset.append(value)
            elif index.attrib["name"] == "chromatogram":
                chromatogram_offset.append(value)
    return spectra_offset, chromatogram_offset


def _get_xml_data(
    filename: Path, spectra_offset: list[int], chromatogram_offset: list[int], index_offset: int, n: int, kind: str
) -> str:
    """Get the xml string associated with a spectrum or chromatogram.

    :param filename: str
    :param spectra_offset: offsets obtained from _build_offset_list
    :param chromatogram_offset: offsets obtained from _build_offset_list
    :param index_offset: offset obtained from _get_index_offset
    :param n: number of spectrum/chromatogram to select
    :param kind: {"spectrum", "chromatogram"}

    Returns
    -------
    str

    """
    if kind == "spectrum":
        lst, other = spectra_offset, chromatogram_offset
        end_tag = "</spectrum>"
    elif kind == "chromatogram":
        lst, other = chromatogram_offset, spectra_offset
        end_tag = "</chromatogram>"
    else:
        raise MZMLError("Kind must be `spectrum` or `chromatogram`")

    start = lst[n]
    # Here we search the closest offset to start such that the complete data is contained in the text
    try:
        end = lst[n + 1]
    except IndexError:
        try:
            end = other[0]
        except IndexError:
            end = index_offset

    if end < start:
        end = index_offset

    with open(filename, "r") as fin:
        fin.seek(start)
        chunk = fin.read(end - start)
    end = chunk.find(end_tag)
    return chunk[: end + len(end_tag)]


def _get_time(element):
    for e in list(element):
        tag = e.tag
        if tag == "scan":
            for ee in list(e):
                accession = ee.attrib.get("accession")
                if accession == _ArrayType.TIME:
                    value = float(ee.attrib.get("value"))
                    units = ee.attrib.get("unitAccession")
                    value = _time_to_seconds(value, units)
                    return value


def _time_to_seconds(value, units):
    if units == _TimeUnits.MINUTES:
        value = value * 60
    return value



from itertools import product

import numpy as np
import pytest

from tidyms2.chem import PTABLE
from tidyms2.chem.envelope_utils import HomoAtomicEnvelopeCache, combine_envelopes


class TestHomoAtomicEnvelopeCache:
    @pytest.fixture(scope="module")
    def cache(self):
        return HomoAtomicEnvelopeCache()

    @pytest.mark.parametrize("symbol,n,length", product(["2H", "31P"], [0, 1, 5], [1, 2, 5]))
    def test_get_envelope_from_isotope(self, cache: HomoAtomicEnvelopeCache, symbol: str, n: int, length: int):
        isotope = PTABLE.get_isotope(symbol)
        actual = cache.get_envelope(isotope, n, length)
        M_expected = np.zeros((1, length))
        M_expected[0, 0] = n * isotope.m
        p_expected = np.zeros((1, length))
        p_expected[0, 0] = 1.0
        assert np.array_equal(actual.M, M_expected)
        assert np.array_equal(actual.p, p_expected)

    @pytest.mark.parametrize("symbol,n,length", product(["23Na", "31P"], [0, 1, 5], [1, 2, 5]))
    def test_get_envelope_from_isotope_and_element_are_equal_for_elements_with_one_isotope(
        self, cache: HomoAtomicEnvelopeCache, symbol: str, n: int, length: int
    ):
        isotope = PTABLE.get_isotope(symbol)
        element = PTABLE.get_element(isotope.z)
        env_isotope = cache.get_envelope(isotope, n, length)
        env_element = cache.get_envelope(element, n, length)
        assert np.array_equal(env_isotope.M, env_element.M)
        assert np.array_equal(env_isotope.p, env_element.p)

    def test_create_envelope_array_from_zero_to_zero(self, cache: HomoAtomicEnvelopeCache):
        symbol = "P"
        element = PTABLE.get_element(symbol)
        actual = cache.make_envelope_array(element, 0, 0, 10)
        expected = cache.get_empty()
        assert np.array_equal(actual.M, expected.M)
        assert np.array_equal(actual.p, expected.p)

    def test_create_envelope_array(self, cache: HomoAtomicEnvelopeCache):
        symbol = "P"
        element = PTABLE.get_element(symbol)
        n_max = 5
        length = 10
        actual = cache.make_envelope_array(element, 0, n_max, length)
        for k in range(n_max):
            expected_row = cache.get_envelope(element, k, length)
            assert np.array_equal(actual.M[k], expected_row.M[0])
            assert np.array_equal(actual.p[k], expected_row.p[0])


@pytest.mark.parametrize("symbol", ["C", "S", "P"])
def test_combine_envelopes(symbol):
    element = PTABLE.get_element(symbol)
    cache = HomoAtomicEnvelopeCache()
    n1 = 5
    env1 = cache.get_envelope(element, n1, length=10)
    n2 = 3
    env2 = cache.get_envelope(element, n2, length=10)
    n = n1 + n2
    actual = combine_envelopes(env1, env2)
    expected = cache.get_envelope(element, n, length=10)
    assert np.allclose(actual.M, expected.M, atol=10e-5)
    assert np.allclose(actual.p, expected.p, atol=10e-5)



"""Common utilities for data simulation."""

from pathlib import Path
from random import choice

from ..core.enums import SampleType
from ..core.models import Sample


def create_sample_list(
    sample_types: list[SampleType],
    type_to_groups: dict[SampleType, list[str]] | None = None,
    n_batches: int = 1,
    start_order: int = 0,
) -> list[Sample]:
    """Create a list of sample models using a list of sample types as template.

    :param sample_types: the type of each simulated sample
    :param type_to_groups: sample groups to assign randomly to each sample. If not provided,
        the sample type is used.
    :param n_batches: the number of batches to simulate. If ``n_batches > 1`` then the sample
        types template is concatenated to build each batch of samples increasing the
        sample order and batch accordingly. Note that sample groups may vary across
        batches for due to the fact that the sample group is randomly sampled from
        `type_to_groups`.
    :param start_order: the start value for order.

    """
    if type_to_groups is None:
        type_to_groups = {x: [x] for x in sample_types}

    samples = list()
    for b in range(n_batches):
        batch = [_create_sample(t, i, b, type_to_groups) for i, t in enumerate(sample_types, start=start_order)]
        samples.extend(batch)
        start_order += len(batch)
    return samples


def _create_sample(type_: SampleType, order: int, batch: int, type_to_groups: dict[SampleType, list[str]]) -> Sample:
    if type_ not in type_to_groups:
        raise ValueError(f"sample type {type_} not included `type_to_groups`.")
    return Sample.model_validate(
        {
            "id": f"sample-{order}",
            "path": Path("."),
            "meta": {
                "group": choice(type_to_groups[type_]),
                "order": order,
                "batch": batch,
                "type": type_,
            },
        }
    )



"""Abstract base operator."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Self

import pydantic

from ..dataflow import ProcessType, check_process_status, update_process_status
from ..enums import MSInstrument, Polarity, SeparationMode


class BaseOperator(ABC, pydantic.BaseModel, Generic[ProcessType]):
    """TidyMS base operator which all other operators inherit from.

    Provides functionality to:
    - set default parameters using instrument type, separation type and polarity.
    - set parameters using a dictionary.
    - get default parameters.

    """

    id: str = ""
    """The Operator id."""

    model_config = pydantic.ConfigDict(validate_assignment=True)

    @abstractmethod
    def get_expected_status_in(self) -> ProcessType:
        """Get the expected sample status before applying the operator."""
        ...

    @abstractmethod
    def get_expected_status_out(self) -> ProcessType:
        """Get the expected sample status after applying the operator."""
        ...

    def check_status(self, status: ProcessType) -> None:
        """Raise an exception if data status is not compatible with operator required status."""
        check_process_status(status, self.get_expected_status_in())

    def update_status(self, status_in: ProcessType):
        """Update the sample process status to the status after applying the operator."""
        return update_process_status(status_in, self.get_expected_status_out())

    @classmethod
    @abstractmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        """Create a new operator with sane defaults for the specified MS instrument, separation mode and polarity.

        :param instrument: The MS instrument used to measure the samples
        :param separation: The analytical method used for separation
        :param polarity: The polarity in which the samples where measured
        :return: A new operator instance

        """
        ...



"""The data flow models define valid workflows."""

from typing import Literal, TypeVar

import pydantic

from .enums import OperatorType
from .exceptions import ProcessStatusError


class BaseProcessStatus(pydantic.BaseModel):
    """Base model to store data storage status."""

    extra: dict[str, bool] = dict()
    """Extra fields added to any model."""

    def list_extras(self) -> list[str]:
        """List all extra fields."""
        return list(self.extra)

    def get_extra(self, field: str) -> bool:
        """Retrieve extra value. Return ``False`` if not found."""
        if field not in self.extra:
            return False
        return self.extra[field]

    def set_extra(self, field: str, value: bool) -> None:
        """Set extra value."""
        self.extra[field] = value


class SampleProcessStatus(BaseProcessStatus):
    """Report sample data process status.

    Each attribute is a flag representing the current status of the data through
    a processing pipeline.

    """

    type: Literal[OperatorType.SAMPLE] = OperatorType.SAMPLE

    roi_extracted: bool = False
    """Flags if ROI were extracted."""

    feature_extracted: bool = False
    """Flags if features were extracted."""

    isotopologue_annotated: bool = False
    """Flags if isotopologue features were annotated."""

    adduct_annotated: bool = False
    """Flags if adducts were annotated."""


class AssayProcessStatus(BaseProcessStatus):
    """Report assay data process status.

    Each attribute is a flag representing the current status of the data through
    a processing pipeline.

    """

    type: Literal[OperatorType.ASSAY] = OperatorType.ASSAY

    adduct_annotated: bool = False
    """Flags if adducts were annotated."""

    feature_matched: bool = False
    """Flags if features where matched"""

    features_group_created: bool = False
    """Flags if feature groups were created"""

    isotopologue_annotated: bool = False
    """Flags if isotopologue features were annotated."""

    missing_imputed: bool = False
    """Flags if missing values were imputed."""


class DataMatrixProcessStatus(BaseProcessStatus):
    """Report assay data process status.

    Each attribute is a flag representing the current status of the data through
    a processing pipeline.

    """

    type: Literal[OperatorType.MATRIX] = OperatorType.MATRIX

    normalized: bool = False
    """Flags if Sample values are normalized."""

    missing_imputed: bool = False
    """Flag if the matrix contain missing values."""


ProcessType = TypeVar("ProcessType", SampleProcessStatus, AssayProcessStatus, DataMatrixProcessStatus)


def check_process_status(actual: ProcessType, reference: ProcessType) -> None:
    """Check if sample status is compatible with a reference status.

    :param actual: The actual sample status
    :param reference: The status used as a reference
    :raises ProcessStatusError: if actual status is not compatible with the reference status

    """
    for field in reference.__class__.model_fields:
        if field == "type" or field == "extra":
            continue

        if not getattr(actual, field) and getattr(reference, field):
            msg = f"Expected {actual.type.value} status {field} to be True."
            raise ProcessStatusError(msg)

    for field in reference.list_extras():
        if reference.get_extra(field) and not actual.get_extra(field):
            msg = f"Expected extra check field {field} in {actual.type.value} to be True."
            raise ProcessStatusError(msg)


def update_process_status(actual: ProcessType, reference: ProcessType) -> None:
    """Update two compatible status using the reference.

    :param actual: the status to update
    :param reference: The status used as a reference

    """
    for field in reference.__class__.model_fields:
        if field == "type" or field == "extra":
            continue
        value = getattr(actual, field) or getattr(reference, field)
        setattr(actual, field, value)

    for extra in reference.list_extras():
        value = reference.get_extra(extra) or actual.get_extra(extra)
        actual.set_extra(extra, value)



"""Utilities to create LC-MS assays."""

from ..annotation.operators import IsotopologueAnnotator
from ..assay import Assay
from ..assay.executors import ParallelSampleProcessor, SequentialSampleProcessor
from ..core.enums import MSInstrument, Polarity, SeparationMode
from ..core.models import MZTrace
from ..storage.memory import OnMemoryAssayStorage, OnMemorySampleStorage
from ..storage.sqlite import SQLiteAssayStorage
from .models import Peak
from .operators import LCFeatureMatcher, LCPeakExtractor, LCTraceBaselineEstimator, LCTraceExtractor


def create_lcms_assay(
    id: str,
    *,
    instrument: MSInstrument | str,
    separation: SeparationMode | str,
    polarity: Polarity | str,
    annotate_isotopologues: bool = True,
    on_disk: bool = False,
    max_workers: int = 1,
    storage_path: str | None = None,
) -> Assay[MZTrace, Peak]:
    """Create a new Assay instance for LC-MS data.

    :param id: the assay name
    :param instrument: the instrument used in the experimental measurements. Used to define operator defaults.
    :param separation: the separation mode used in the experimental measurements. Used to define operator defaults.
    :param polarity: the instrument polarity. Used to define operator defaults.
    :param instrument: the instrument used for the assay. Used to define operator defaults.
    :param annotate_isotopologues: If set to ``True`` and isotopologue annotation step is included in
        the sample pipeline.
    :param on_disk: store assay results on disk to reduce memory consumption. Recommended for large datasets.
    :param storage_path: path to the DB file to store assay data. Only used if `on_disk` is set to ``True``.

    """
    if isinstance(instrument, str):
        instrument = MSInstrument(instrument)

    if isinstance(separation, str):
        separation = SeparationMode(separation)

    if isinstance(polarity, str):
        polarity = Polarity(polarity)

    if on_disk:
        host = storage_path or f"{id}.sqlite"
        storage = SQLiteAssayStorage(id, host, MZTrace, Peak)
    else:
        storage = OnMemoryAssayStorage(id, MZTrace, Peak)

    if max_workers == 1:
        executor = SequentialSampleProcessor()
    else:
        executor = ParallelSampleProcessor(max_workers=max_workers)

    assay = Assay(
        id=id,
        assay_storage=storage,
        sample_processor=executor,
        sample_storage_type=OnMemorySampleStorage,
    )

    sample_ops = list()
    sample_ops.append(LCTraceExtractor.from_defaults(instrument, separation, polarity))
    sample_ops.append(LCTraceBaselineEstimator.from_defaults(instrument, separation, polarity))
    sample_ops.append(LCPeakExtractor.from_defaults(instrument, separation, polarity))
    if annotate_isotopologues:
        sample_ops.append(IsotopologueAnnotator.from_defaults(instrument, separation, polarity))

    for op in sample_ops:
        op.id = f"{assay.id}__{op.__class__.__name__}"
        assay.pipes.sample.add_operator(op)

    assay.pipes.assay.add_operator(LCFeatureMatcher.from_defaults(instrument, separation, polarity))

    for op in assay.pipes.assay.operators:
        op.id = f"{assay.id}-{op.__class__.__name__}"

    return assay



"""Container class for assay features descriptors and annotations."""

from typing import Iterable
from uuid import UUID

from ..core.exceptions import FeatureGroupNotFound, FeatureNotFound, SampleNotFound
from ..core.models import Annotation, AnnotationPatch, DescriptorPatch, FeatureGroup


class FeatureTable:
    """Store Feature annotations and descriptors in long, i.e., one feature record per row, format."""

    def __init__(self):
        self._descriptors: dict[str, list[float]] = dict()
        self._annotations: list[Annotation] = list()
        self._fill_values: dict[str, dict[int, float]] = dict()
        self._feature_groups: dict[int, FeatureGroup] = dict()

        self._sample_to_features: dict[str, set[UUID]] = dict()
        self._group_to_features: dict[int, set[UUID]] = dict()

        # maintains feature index in the snapshot for fast access
        self._feature_to_index: dict[UUID, int] = dict()
        self._current_index = 0

    def add_descriptors(self, descriptors: list[dict[str, float]], annotations: list[Annotation]):
        """Add annotations and descriptors from a sample."""
        for descriptor in descriptors:
            for name, value in descriptor.items():
                descriptor_list = self._descriptors.setdefault(name, list())
                descriptor_list.append(value)
        self._annotations.extend(annotations)

        for k, ann in enumerate(annotations, start=self._current_index):
            sample_set = self._sample_to_features.setdefault(ann.sample_id, set())
            sample_set.add(ann.id)
            if ann.group > -1:
                group_set = self._group_to_features.setdefault(ann.group, set())
                group_set.add(ann.id)

            self._feature_to_index[ann.id] = k

        self._current_index += len(descriptors)

    def add_feature_groups(self, *feature_groups: FeatureGroup) -> None:
        """Add feature groups to the snapshot."""
        for group in feature_groups:
            self._feature_groups[group.group] = group

    def fetch_feature_groups(self) -> list[FeatureGroup]:
        """Fetch feature groups from the snapshot."""
        return [x.model_copy(deep=True) for x in self._feature_groups.values()]

    def has_feature_group(self, group: int) -> bool:
        """Check if a group with the provided id is stored in the assay."""
        return group in self._feature_groups

    def fetch_annotations(self, sample_id: str | None = None, copy: bool = False) -> list[Annotation]:
        """Create a list feature annotations.

        :param sample_id: If provided, only include annotations from this sample
        """
        if sample_id is None:
            result = [x for x in self._annotations]
        elif sample_id not in self._sample_to_features:
            raise SampleNotFound(sample_id)
        else:
            indices = [self._feature_to_index[x] for x in self._sample_to_features[sample_id]]
            result = [self._annotations[x] for x in indices]

        if copy:
            result = [x.model_copy() for x in result]

        return result

    def fetch_descriptors(
        self, descriptors: Iterable[str] | None = None, sample_id: str | None = None, copy: bool = False
    ) -> dict[str, list[float]]:
        """Fetch descriptors from the snapshot."""
        if descriptors is None:
            descriptors = list(self._descriptors)
        if sample_id is None and copy:
            result = {x: self._descriptors[x].copy() for x in descriptors}
        elif sample_id is None:
            result = {x: self._descriptors[x] for x in descriptors}
        elif sample_id not in self._sample_to_features:
            raise SampleNotFound(sample_id)
        else:
            indices = [self._feature_to_index[x] for x in self._sample_to_features[sample_id]]
            result = dict()
            for name in descriptors:
                values = self._descriptors[name]
                result[name] = [values[x] for x in indices]
        return result

    def fetch_fill_values(self, copy: bool = False) -> dict[str, dict[int, float]]:
        """Fetch snapshot fill values."""
        if copy:
            return {k: v.copy() for k, v in self._fill_values.items()}
        return self._fill_values

    def add_fill_values(self, fill_values: dict[str, dict[int, float]]) -> None:
        """Add missing values to the snapshot."""
        for sample_id, features_fill in fill_values.items():
            for feature_group, value in features_fill.items():
                sample_fill = self._fill_values.setdefault(sample_id, dict())
                sample_fill[feature_group] = value

    def get_sample_id(self, feature_id: UUID) -> str:
        """Retrieve the sample id of a feature."""
        index = self._feature_to_index[feature_id]
        return self._annotations[index].sample_id

    def get_ids_by_group(self, group: int) -> list[UUID]:
        """Retrieve all feature ids associated with a feature group."""
        if group not in self._group_to_features:
            raise FeatureGroupNotFound(group)
        return [x for x in self._group_to_features[group]]

    def has_feature(self, feature_id: UUID) -> bool:
        """Check if a feature is in the snapshot."""
        return feature_id in self._feature_to_index

    def patch_annotation(self, *patches: AnnotationPatch) -> None:
        """Apply patches to annotations."""
        for p in patches:
            if not self.has_feature(p.id):
                raise FeatureNotFound(p.id)

        for p in patches:
            index = self._feature_to_index[p.id]
            ann = self._annotations[index]
            setattr(ann, p.field, p.value)
            self._update_group_to_features(p)

    def list_feature_groups(self) -> list[int]:
        """List all feature groups stored in the assay."""
        return list(self._feature_groups)

    def patch_descriptors(self, *patches: DescriptorPatch) -> None:
        """Apply patches to descriptors."""
        for p in patches:
            if not self.has_feature(p.id):
                raise FeatureNotFound(p.id)

        for p in patches:
            index = self._feature_to_index[p.id]
            self._descriptors[p.descriptor][index] = p.value

    def _update_group_to_features(self, patch: AnnotationPatch):
        """Update the group to feature id mapping when the group annotation is update."""
        if patch.field == "group" and patch.value > -1:
            group_set = self._group_to_features.setdefault(patch.value, set())
            group_set.add(patch.id)
